import linecache
import logging
import os.path
import sys


def get_exception_source():
    """Returns full file path, file name, line number, function name, and line contents
    causing the last exception."""
    _, _, tb = sys.exc_info()
    while tb.tb_next:
        tb = tb.tb_next
    f = tb.tb_frame
    lineno = tb.tb_lineno
    co = f.f_code
    filefullpath = co.co_filename
    filename = os.path.basename(filefullpath)
    name = co.co_name
    linecache.checkcache(filefullpath)
    line = linecache.getline(filefullpath, lineno, f.f_globals)
    if line:
        line = line.strip()
    else:
        line = None
    return filefullpath, filename, lineno, name, line


class CallbackHandler(logging.Handler):
    """This handler sends logrecords to a callback function."""

    def __init__(self, callback):
        """Set up a handler instance, record the callback function."""
        super().__init__()
        self._callback = callback

    def prepare(self, record):
        # Function taken from Python 3.6 QueueHandler
        """
        Prepares a record for queuing. The object returned by this method is
        enqueued.
        The base implementation formats the record to merge the message
        and arguments, and removes unpickleable items from the record
        in-place.
        You might want to override this method if you want to convert
        the record to a dict or JSON string, or send a modified copy
        of the record while leaving the original intact.
        """
        # The format operation gets traceback text into record.exc_text
        # (if there's exception data), and also puts the message into
        # record.message. We can then use this to replace the original
        # msg + args, as these might be unpickleable. We also zap the
        # exc_info attribute, as it's no longer needed and, if not None,
        # will typically not be pickleable.
        self.format(record)
        record.msg = record.message
        record.args = None
        record.exc_info = None
        return record

    def emit(self, record):
        """Send a LogRecord to the callback function, after preparing it
        for serialization."""
        try:
            self._callback(self.prepare(record))
        except Exception:
            self.handleError(record)



import random
import time

from workflows.services.common_service import CommonService


class SampleTxn(CommonService):
    """An example service building on top of the workflow.services architecture,
    demonstrating how this architecture can be used.
    This service consumes a messages off one queue and places it into another.
    Transactions are used to guarantee correct message handling.
    The service is deliberately unreliable and prone to failure."""

    # Human readable service name
    _service_name = "Transaction sample"

    def initializing(self):
        """Subscribe to a channel. Received messages must be acknowledged."""
        self.subid = self._transport.subscribe(
            "transient.transaction", self.receive_message, acknowledgement=True
        )

    @staticmethod
    def crashpoint():
        """Return true if the service should malfunction at this point."""
        # Probability of not crashing is 90%
        return random.uniform(0, 1) > 0.90

    def receive_message(self, header, message):
        """Receive a message"""

        print("=== Receive ===")
        print(header)
        print(message)

        print("MsgID: {}".format(header["message-id"]))
        assert header["message-id"]

        txn = self._transport.transaction_begin()
        print(" 1. Txn: {}".format(str(txn)))
        if self.crashpoint():
            self._transport.transaction_abort(txn)
            print("---  Abort  ---")
            return

        self._transport.ack(header["message-id"], self.subid, transaction=txn)
        print(" 2. Ack")
        if self.crashpoint():
            self._transport.transaction_abort(txn)
            print("---  Abort  ---")
            return

        self._transport.send("transient.destination", message, transaction=txn)
        print(" 3. Send")

        if self.crashpoint():
            self._transport.transaction_abort(txn)
            print("---  Abort  ---")
            return

        self._transport.transaction_commit(txn)
        print(" 4. Commit")
        print("===  Done   ===")


class SampleTxnProducer(CommonService):
    """An example service building on top of the workflow.services architecture,
    demonstrating how this architecture can be used.
    This service generates messages for the Transaction example."""

    # Human readable service name
    _service_name = "TXN Message Producer"

    counter = 0

    def initializing(self):
        """Service initialization. This function is run before any commands are
        received from the frontend. This is the place to request channel
        subscriptions with the messaging layer, and register callbacks.
        This function can be overridden by specific service implementations."""
        self._register_idle(3, self.create_message)

    def create_message(self):
        """Create and send a unique message for this service."""
        self.counter += 1
        self._transport.send(
            "transient.transaction",
            "TXMessage #%d\n++++++++Produced@ %f"
            % (self.counter, (time.time() % 1000) * 1000),
        )
        self.log.info("Created message %d", self.counter)



from unittest import mock

import pytest
import workflows
import workflows.recipe


def generate_recipes():
    """Generate two recipe objects for testing."""

    class RecipeA(workflows.recipe.Recipe):
        """Testing recipe class by overriding the 'recipe' property."""

        recipe = {
            1: {
                "service": "A service",
                "queue": "some.queue.{first}",
                "output": [2],
                "error": 2,
            },
            2: {"service": "B service", "queue": "another.queue.{name}"},
            "start": [(1, {})],
            "error": [2],
        }

    recipe_b = {
        1: {"service": "A service", "queue": "some.queue.{first}", "output": 2},
        2: {"service": "C service", "queue": "third.queue"},
        "start": [(1, {})],
    }

    return RecipeA(), workflows.recipe.Recipe(recipe_b)


def test_can_generate_recipe_objects():
    """Test generation of recipes."""
    A, B = generate_recipes()

    # Check that both recipies are valid
    A.validate()
    B.validate()
    assert len(B.recipe) == 3


def test_equality_and_inequality_operator():
    """Check that recipe objects can be compared with recipe objects,
    string and dictionary representations of recipe objects."""
    A, _ = generate_recipes()
    B, _ = generate_recipes()

    assert A == B
    assert A == B.serialize()
    assert A == B.pretty()
    assert A == B.recipe

    # This is a semi-mangled recipe, containing a string pointer instead of an integer pointer
    # This can't happen with a recipe object, but can happen with dictionary representation.
    # They are still equivalent.
    B.recipe["1"] = B.recipe[1]
    del B.recipe[1]
    assert A == B.recipe

    # This is another semi-mangled recipe, containing a single integer pointer instead of a list as a default output.
    # Again this can happen with dictionary representation, and the recipes are still equivalent.
    B.recipe["1"]["output"] = 2
    assert A == B.recipe

    # This is a different (invalid) recipe.
    del B.recipe["1"]
    assert A != B
    assert A != B.serialize()
    assert A != B.recipe


def test_in_operator():
    """Check that the 'in' operator is properly supported by the recipe object."""
    A, B = generate_recipes()

    assert 1 in A
    assert 2 in A
    assert 3 not in A
    assert "start" in A
    assert "error" in A

    assert 1 in B
    assert 2 in B
    assert 3 not in B
    assert "start" in B
    assert "error" not in B


def test_serializing_and_deserializing_recipes():
    """Test generation of recipes."""
    A, B = generate_recipes()

    # Check that both recipies are valid
    assert A.deserialize(A.serialize()) == A.recipe
    assert B.deserialize(B.serialize()) == B.recipe


def test_validate_tests_for_empty_recipe():
    """Validating a recipe that has not been defined must throw an error."""
    A, _ = generate_recipes()
    A.recipe = None
    with pytest.raises(workflows.Error):
        A.validate()


def test_validate_tests_for_invalid_nodes():
    """Check that the recipe contains only numeric nodes, a non-empty 'start' node, and an optional 'error' node."""
    A, _ = generate_recipes()
    A.recipe["xnode"] = None
    with pytest.raises(workflows.Error) as excinfo:
        A.validate()
    assert "xnode" in str(excinfo.value)

    A, _ = generate_recipes()
    del A.recipe["start"]
    with pytest.raises(workflows.Error) as excinfo:
        A.validate()
    assert "start" in str(excinfo.value)

    A.recipe["start"] = []
    with pytest.raises(workflows.Error) as excinfo:
        A.validate()
    assert "start" in str(excinfo.value)

    A.recipe["start"] = [("something",)]
    with pytest.raises(workflows.Error) as excinfo:
        A.validate()
    assert "start" in str(excinfo.value)

    A.recipe["start"] = [(1, "something"), "banana"]
    with pytest.raises(workflows.Error) as excinfo:
        A.validate()
    assert "start" in str(excinfo.value)

    A.recipe["start"] = [(1, "something"), ("start", "banana")]
    with pytest.raises(workflows.Error) as excinfo:
        A.validate()
    assert "start" in str(excinfo.value)

    A, _ = generate_recipes()
    del A.recipe["error"]
    A.validate()

    A.recipe["error"] = []
    A.validate()

    A.recipe["error"] = 1
    A.validate()

    A.recipe["error"] = [1, 2]
    A.validate()

    A.recipe["error"] = "something"
    with pytest.raises(workflows.Error) as excinfo:
        A.validate()
    assert "error" in str(excinfo.value)


def test_validate_tests_for_invalid_links():
    """Check that the nodes in recipes have valid output/error links to other nodes."""

    # Part 1: Check outgoing links from start/error node:
    A, _ = generate_recipes()
    A.recipe["start"] = [("asdf", None)]
    with pytest.raises(workflows.Error):
        A.validate()

    A.recipe["start"] = [(1, None), (2, None)]
    A.validate()

    A.recipe["start"] = [(1, None), (99, None)]
    with pytest.raises(workflows.Error):
        A.validate()

    A, _ = generate_recipes()
    A.recipe["error"] = ["start", 2]
    with pytest.raises(workflows.Error) as excinfo:
        A.validate()
    assert "error" in str(excinfo.value)

    A.recipe["error"] = "error"
    with pytest.raises(workflows.Error) as excinfo:
        A.validate()
    assert "error" in str(excinfo.value)

    A.recipe["error"] = 99
    with pytest.raises(workflows.Error):
        A.validate()

    # Part 2 & 3: Check outgoing links from other nodes (output and error)
    for outgoing in ("output", "error"):
        A, _ = generate_recipes()
        A.recipe[1][outgoing] = "asdf"
        with pytest.raises(workflows.Error):
            A.validate()

        A.recipe[1][outgoing] = 99
        with pytest.raises(workflows.Error):
            A.validate()

        A.recipe[1][outgoing] = [2, 99]
        with pytest.raises(workflows.Error):
            A.validate()

        A.recipe[1][outgoing] = [2, "banana"]
        with pytest.raises(workflows.Error):
            A.validate()

        A.recipe[1][outgoing] = {"all": 2, "some": [2, 2]}
        A.validate()

        A.recipe[1][outgoing] = {"all": 2, "some": [2, 99]}
        with pytest.raises(workflows.Error):
            A.validate()

    # Part 4: Check for unreferenced nodes
    A, _ = generate_recipes()
    A.recipe[99] = {}
    with pytest.raises(workflows.Error) as excinfo:
        A.validate()
    assert "99" in str(excinfo.value)


def test_validate_tests_for_cycles():
    """Check that validation detects cycles in recipes.  Recipes must be acyclical."""
    A, _ = generate_recipes()
    A.recipe[2]["output"] = 1
    with pytest.raises(workflows.Error) as excinfo:
        A.validate()
    assert "cycle" in str(excinfo.value)

    A, _ = generate_recipes()
    A.recipe[2]["output"] = 2
    with pytest.raises(workflows.Error) as excinfo:
        A.validate()
    assert "cycle" in str(excinfo.value)

    A, _ = generate_recipes()
    A.recipe[2]["output"] = [1, 2]
    with pytest.raises(workflows.Error) as excinfo:
        A.validate()
    assert "cycle" in str(excinfo.value)


def test_replacing_parameters_in_recipe():
    """Recipe may contain placeholders that should be replaced with actual values by running apply_parameters."""
    A, _ = generate_recipes()

    replacements = {"name": "replacement"}

    A.apply_parameters(replacements)

    assert A.recipe[1]["queue"] == "some.queue.{first}"
    assert A.recipe[2]["queue"] == "another.queue.replacement"


def test_replacing_parameters_in_recipe_should_keep_unknown_parameter_strings():
    """Recipe may contain placeholders that can't be resolved. These should stay as they are."""
    A = workflows.recipe.Recipe(
        {
            1: {"service": "A service", "queue": "some.queue.{first}", "output": 2},
            2: {
                "service": "C service",
                "queue": "another.queue.{unknownparameter[key][subkey]}",
            },
            "start": [(1, {})],
        }
    )

    replacements = {"name": "replacement"}

    A.apply_parameters(replacements)

    assert A.recipe[1]["queue"] == "some.queue.{first}"
    assert A.recipe[2]["queue"] == "another.queue.{unknownparameter[key][subkey]}"


def test_replacing_parameters_in_recipe_with_datastructures():
    """Recipe may contain placeholders that should be replaced with actual data structures by running apply_parameters."""
    A, _ = generate_recipes()

    replacements = {
        "name": "replacement",
        "some-list": [1, 2, 3],
        "nested": {
            "some-dictionary": {
                "number": 3,
                "oddlist": [1, 3, 5],
                "text": "string",
                "dict": {"key": "value"},
            }
        },
    }

    A.recipe[2]["list"] = "{$REPLACE:some-list}"
    A.recipe[2]["dictionary"] = "{$REPLACE:nested}"
    A.recipe[2]["deep-dictionary"] = "{$REPLACE:nested[some-dictionary]}"
    A.recipe[2]["undefined"] = "{$REPLACE:undefined}"
    A.recipe[2]["undefined-deep-dictionary"] = "{$REPLACE:nested[undefined-dictionary]}"
    A.apply_parameters(replacements)

    assert A.recipe[2]["list"] == replacements["some-list"]
    assert (
        A.recipe[2]["list"] is not replacements["some-list"]
    ), "same list returned, not copied"
    assert A.recipe[2]["dictionary"] == replacements["nested"]
    assert (
        A.recipe[2]["dictionary"] is not replacements["nested"]
    ), "same dictionary returned, not copied"
    assert (
        A.recipe[2]["dictionary"]["some-dictionary"]
        is not replacements["nested"]["some-dictionary"]
    ), "same dictionary returned, not deep-copied"
    assert A.recipe[2]["deep-dictionary"] == replacements["nested"]["some-dictionary"]
    assert (
        A.recipe[2]["deep-dictionary"] is not replacements["nested"]["some-dictionary"]
    )
    assert A.recipe[2]["undefined"] is None
    assert A.recipe[2]["undefined-deep-dictionary"] is None


def test_merging_recipes():
    """Test recipes can be merged and merging results in a valid minimal DAG."""
    A, B = generate_recipes()

    # Merging empty recipes returns original recipe
    C = A.merge(None)
    assert A == C

    C = workflows.recipe.Recipe().merge(A)
    assert A == C

    C = A.merge(workflows.recipe.Recipe())
    assert A == C

    C = A.merge(B)

    # Merge function should not modify original recipes
    assert A, B == generate_recipes()

    # Result will have 6 nodes: start, error, A1, A2, B1, B2
    assert len(C.recipe) == 6

    # Start node contains two different pointers to 'A service'
    assert "start" in C.recipe
    assert len(C.recipe["start"]) == 2
    assert C.recipe["start"][0] == (1, {})
    assert C.recipe["start"][1] == (mock.ANY, {})
    assert C.recipe["start"][0][0] != C.recipe["start"][1][0]
    assert C.recipe[C.recipe["start"][0][0]]["service"] == "A service"
    assert C.recipe[C.recipe["start"][1][0]]["service"] == "A service"

    # Error node points to 'B service'
    assert "error" in C.recipe
    assert len(C.recipe["error"]) == 1
    assert C.recipe[C.recipe["error"][0]]["service"] == "B service"

    # There is a 'C service'
    assert any(
        map(
            lambda x: (isinstance(x, dict) and x.get("service") == "C service"),
            C.recipe.values(),
        )
    )



import workflows


def test_workflows_version_returns_sensible_value():
    """The version() function should return something
    resembling a version number."""
    assert workflows.version().index(".")



from unittest import mock
import workflows.frontend.utilization
from workflows.services.common_service import Status


def about(value, tolerance):
    """Create an object that can be compared against a number and allows a tolerance."""

    class Comparator:
        """A helper class to compare against a value with a tolerance."""

        def __le__(self, other):
            return other >= value - tolerance

        def __eq__(self, other):
            return value - tolerance <= other <= value + tolerance

        def __ge__(self, other):
            return other <= value + tolerance

        def __ne__(self, other):
            return not self.__eq__(other)

        def __repr__(self):
            return "<{} +- {}>".format(str(value), str(tolerance))

    return Comparator()


def test_near_equality_helper_class():
    """Quick test of the above helper class."""
    three = about(3, 0.11)
    assert 2 <= three
    assert 2 != three
    assert 2.8 != three
    assert 2.9 == three
    assert 3.0 == three
    assert 3 == three
    assert 3.1 == three
    assert 3.2 != three
    assert 4 != three
    assert 4 >= three


def test_get_empty_statistics_report():
    """Create a UtilizationStatistics object and get the initial report. It should report being in the 'NEW' status 100% of the time."""
    stat = workflows.frontend.utilization.UtilizationStatistics(summation_period=10)
    assert stat.report() == {Status.NEW.intval: 1.0}


@mock.patch("workflows.frontend.utilization.time")
def test_statistics_report_contains_correctly_aggregated_information(t):
    """Create a UtilizationStatistics object and feed in various status changes. Check that the aggregated reports correspond to expected values."""
    t.time.return_value = 100000
    stat = workflows.frontend.utilization.UtilizationStatistics(summation_period=10)
    t.time.return_value = 100005
    assert stat.report() == {Status.NEW.intval: 1.0}
    stat.update_status(Status.IDLE.intval)
    t.time.return_value = 100010
    assert stat.report() == {
        Status.NEW.intval: about(0.5, 0.01),
        Status.IDLE.intval: about(0.5, 0.01),
    }
    stat.update_status(127)
    t.time.return_value = 100012
    assert stat.report() == {
        Status.NEW.intval: about(0.3, 0.01),
        Status.IDLE.intval: about(0.5, 0.01),
        127: about(0.2, 0.01),
    }
    stat.update_status(Status.IDLE.intval)
    t.time.return_value = 100013
    assert stat.report() == {
        Status.NEW.intval: about(0.2, 0.01),
        Status.IDLE.intval: about(0.6, 0.01),
        127: about(0.2, 0.01),
    }
    stat.update_status(128)
    t.time.return_value = 100016
    assert stat.report() == {
        Status.IDLE.intval: about(0.5, 0.01),
        127: about(0.2, 0.01),
        128: about(0.3, 0.01),
    }
    t.time.return_value = 100017
    assert stat.report() == {
        Status.IDLE.intval: about(0.4, 0.01),
        127: about(0.2, 0.01),
        128: about(0.4, 0.01),
    }
    t.time.return_value = 100018
    assert stat.report() == {
        Status.IDLE.intval: about(0.3, 0.01),
        127: about(0.2, 0.01),
        128: about(0.5, 0.01),
    }
    t.time.return_value = 100019
    assert stat.report() == {
        Status.IDLE.intval: about(0.2, 0.01),
        127: about(0.2, 0.01),
        128: about(0.6, 0.01),
    }
    t.time.return_value = 100020
    assert stat.report() == {
        Status.IDLE.intval: about(0.1, 0.01),
        127: about(0.2, 0.01),
        128: about(0.7, 0.01),
    }
    t.time.return_value = 100021
    assert stat.report() == {
        Status.IDLE.intval: about(0.1, 0.01),
        127: about(0.1, 0.01),
        128: about(0.8, 0.01),
    }
    t.time.return_value = 100022
    assert stat.report() == {
        Status.IDLE.intval: about(0.1, 0.01),
        127: about(0.0, 0.01),
        128: about(0.9, 0.01),
    }
    t.time.return_value = 100022.001
    assert stat.report() == {
        Status.IDLE.intval: about(0.1, 0.01),
        128: about(0.9, 0.01),
    }
    t.time.return_value = 100023
    assert stat.report() == {
        Status.IDLE.intval: about(0.0, 0.01),
        128: about(1.0, 0.01),
    }
    t.time.return_value = 100023.001
    assert stat.report() == {128: 1.0}



from unittest import mock

import pytest
import workflows
from workflows.recipe import Recipe
from workflows.recipe.wrapper import RecipeWrapper


def generate_recipe_message():
    """Helper function for tests."""
    message = {
        "recipe": {
            1: {
                "service": "A service",
                "queue": "service.one",
                "output": {"somewhere": [2], "multi-output": [2, 3, 4]},
                "parameter": "some {placeholder} to test replacement",
                "error": 2,
            },
            2: {"service": "service 2", "queue": "queue.two", "output": [3]},
            3: {"service": "service 3", "topic": "topic.three", "transport-delay": 300},
            4: {"service": "service 4", "queue": "queue.four", "transport-delay": 300},
            "start": [(1, {"parameter": mock.sentinel.initial_parameter})],
            "error": [2],
        },
        "recipe-pointer": 1,
        "recipe-path": [],
        "environment": {
            "ID": mock.sentinel.GUID,
            "source": mock.sentinel.source,
            "timestamp": mock.sentinel.timestamp,
        },
        "payload": mock.sentinel.payload,
    }
    return message


def test_recipe_wrapper_instantiated_from_message():
    """A RecipeWrapper built from a message must parse the contained recipe,
    pointers, etc."""
    m = generate_recipe_message()

    rw = RecipeWrapper(message=m)

    assert rw.recipe == Recipe(m["recipe"])
    assert rw.recipe_pointer == m["recipe-pointer"]
    assert rw.recipe_step == m["recipe"][m["recipe-pointer"]]
    assert rw.recipe_path == m["recipe-path"]
    assert rw.environment == m["environment"]
    assert rw.payload == mock.sentinel.payload


def test_recipe_wrapper_instantiated_from_message_with_string_conversion():
    """A RecipeWrapper built from a message must be able to parse numeric strings."""
    m = generate_recipe_message()
    original_pointer = m["recipe-pointer"]
    m["recipe-pointer"] = str(original_pointer)

    rw = RecipeWrapper(message=m)

    assert rw.recipe_pointer == original_pointer
    assert rw.recipe_step == m["recipe"][original_pointer]


def test_recipe_wrapper_instantiated_from_recipe():
    """A RecipeWrapper built from a recipe will contain the recipe, but no
    pointers."""
    r = generate_recipe_message()["recipe"]

    rw = RecipeWrapper(recipe=r)

    assert rw.recipe == Recipe(r)
    assert rw.recipe_pointer is None
    assert rw.recipe_step is None
    assert rw.recipe_path == []
    assert rw.environment == {}
    assert rw.payload is None

    # Constructor can also accept Recipe objects

    r = Recipe(r)
    rw = RecipeWrapper(recipe=r)

    assert rw.recipe == r
    assert rw.recipe_pointer is None
    assert rw.recipe_step is None
    assert rw.recipe_path == []
    assert rw.environment == {}
    assert rw.payload is None


def test_recipe_wrapper_empty_constructor_fails():
    """A RecipeWrapper must be built from either a recipe or a message containing
    a recipe. Otherwise there is nothing to wrap."""
    with pytest.raises(ValueError):
        RecipeWrapper()


def test_downstream_message_sending_via_recipewrapper_with_named_outputs():
    """The idea of the RecipeWrapper is that it makes it easy to send messages from
    one service to another, including instructions on how to handle intermediate
    results at each step. This tests the send_to function which embeds a message
    payload inside a data structure containing all the extra processing info.
    """
    m = generate_recipe_message()
    t = mock.create_autospec(workflows.transport.common_transport.CommonTransport)

    def downstream_message(dest, payload):
        """Helper function to generate expected message contents for downstream
        recipients."""
        ds_message = generate_recipe_message()
        ds_message["recipe-pointer"] = dest
        ds_message["recipe-path"] = [1]
        ds_message["payload"] = payload
        return ds_message

    rw = RecipeWrapper(message=m, transport=t)
    assert rw.transport == t
    assert t.method_calls == []
    t.reset_mock()  # magic call may have been recorded

    # Messages sent to undefined outputs are discarded
    rw.send_to("unknown", mock.sentinel.message_text)
    assert t.mock_calls == []

    # Messages sent to defined outputs are sent
    rw.send_to("somewhere", mock.sentinel.message_text)
    expected = [
        mock.call.send(
            m["recipe"][2]["queue"],
            downstream_message(2, mock.sentinel.message_text),
            headers={"workflows-recipe": True},
        )
    ]
    assert t.mock_calls == expected
    t.reset_mock()

    # Messages sent to outputs with multiple connections are sent to all of them
    rw.send_to(
        "multi-output",
        mock.sentinel.another_message_text,
        transaction=mock.sentinel.txn,
    )
    expected = []
    expected.append(
        mock.call.send(
            m["recipe"][2]["queue"],
            downstream_message(2, mock.sentinel.another_message_text),
            headers={"workflows-recipe": True},
            transaction=mock.sentinel.txn,
        )
    )
    expected.append(
        mock.call.broadcast(
            m["recipe"][3]["topic"],
            downstream_message(3, mock.sentinel.another_message_text),
            headers={"workflows-recipe": True},
            transaction=mock.sentinel.txn,
            delay=300,
        )
    )
    expected.append(
        mock.call.send(
            m["recipe"][4]["queue"],
            downstream_message(4, mock.sentinel.another_message_text),
            headers={"workflows-recipe": True},
            transaction=mock.sentinel.txn,
            delay=300,
        )
    )
    assert t.mock_calls == expected
    t.reset_mock()

    # Messages sent to unnamed output are discarded
    rw.send(mock.sentinel.unnamed_output)
    assert t.mock_calls == []

    # Messages sent to unnamed output are sent when a default output channel is set
    rw.set_default_channel("somewhere")
    rw.send(mock.sentinel.unnamed_output_with_default_name)
    expected = [
        mock.call.send(
            m["recipe"][2]["queue"],
            downstream_message(2, mock.sentinel.unnamed_output_with_default_name),
            headers={"workflows-recipe": True},
        )
    ]
    assert t.mock_calls == expected


def test_downstream_message_sending_via_recipewrapper_with_unnamed_output():
    """Test sending messages via the RecipeWrapper when the current step
    does not have named outputs, or any output at all.
    """

    def downstream_message(dest, payload, path=()):
        """Helper function to generate expected message contents for downstream
        recipients."""
        ds_message = generate_recipe_message()
        ds_message["recipe-pointer"] = dest
        ds_message["recipe-path"] = [1] + list(path)
        ds_message["payload"] = payload
        return ds_message

    m = downstream_message(2, mock.sentinel.payload)
    t = mock.create_autospec(workflows.transport.common_transport.CommonTransport)
    rw = RecipeWrapper(message=m, transport=t)
    t.reset_mock()

    # When only an unnamed output is specified all named outputs are discarded.
    rw.send_to("output", mock.sentinel.test_unnamed_output)
    assert t.mock_calls == []

    # The send() method can deal with unnamed outputs.
    rw.send(mock.sentinel.test_default_output)
    expected = [
        mock.call.broadcast(
            m["recipe"][3]["topic"],
            downstream_message(3, mock.sentinel.test_default_output, path=(2,)),
            headers={"workflows-recipe": True},
            delay=300,
        )
    ]
    assert t.mock_calls == expected
    t.reset_mock()

    # The send_to() method can also deal with unnamed outputs once a default output has been defined.
    rw.set_default_channel("output")
    rw.send_to("output", mock.sentinel.test_default_named_output)
    expected = [
        mock.call.broadcast(
            m["recipe"][3]["topic"],
            downstream_message(3, mock.sentinel.test_default_named_output, path=(2,)),
            headers={"workflows-recipe": True},
            delay=300,
        )
    ]
    assert t.mock_calls == expected


def test_checkpointing_via_recipewrapper():
    """Test resending modified message to the current destination (checkpointing)."""
    m = generate_recipe_message()
    m["recipe-pointer"] = 3
    # Use a step with transport-delay.
    # Checkpointing should ignore default delay settings.

    t = mock.create_autospec(workflows.transport.common_transport.CommonTransport)

    def downstream_message(dest, payload):
        """Helper function to generate expected message contents for downstream
        recipients."""
        ds_message = generate_recipe_message()
        ds_message["recipe-pointer"] = dest
        ds_message["recipe-path"] = []
        ds_message["payload"] = payload
        return ds_message

    # Can't checkpoint on a recipe without a pointer
    rw = RecipeWrapper(recipe=m["recipe"], transport=mock.Mock())
    with pytest.raises(ValueError):
        rw.checkpoint(mock.sentinel.checkpoint, transaction=mock.sentinel.txn)

    rw = RecipeWrapper(message=m, transport=t)
    # Test checkpointing. Delay should be 0.
    rw.checkpoint(mock.sentinel.checkpoint, transaction=mock.sentinel.txn)
    expected = [
        mock.call.broadcast(
            m["recipe"][3]["topic"],
            downstream_message(3, mock.sentinel.checkpoint),
            headers={"workflows-recipe": True},
            delay=0,
            transaction=mock.sentinel.txn,
        )
    ]
    assert t.mock_calls == expected
    t.reset_mock()

    # Test delay override
    rw.checkpoint(mock.sentinel.checkpoint, transaction=mock.sentinel.txn, delay=150)
    expected = [
        mock.call.broadcast(
            m["recipe"][3]["topic"],
            downstream_message(3, mock.sentinel.checkpoint),
            headers={"workflows-recipe": True},
            delay=150,
            transaction=mock.sentinel.txn,
        )
    ]
    assert t.mock_calls == expected


def test_start_command_via_recipewrapper():
    """Test triggering the start of a recipe."""
    m = generate_recipe_message()
    r = m["recipe"]
    t = mock.create_autospec(workflows.transport.common_transport.CommonTransport)

    def downstream_message(dest, payload):
        """Helper function to generate expected message contents for downstream
        recipients."""
        ds_message = generate_recipe_message()
        ds_message["recipe-pointer"] = dest
        ds_message["recipe-path"] = []
        ds_message["payload"] = payload
        ds_message["environment"] = mock.sentinel.environment
        return ds_message

    # Test on already started recipe
    rw = RecipeWrapper(message=m, transport=t)
    with pytest.raises(ValueError):
        rw.start(transaction=mock.sentinel.txn)
    t.reset_mock()

    rw = RecipeWrapper(recipe=r, transport=t)
    rw.environment = mock.sentinel.environment
    assert rw.transport == t
    assert t.method_calls == []
    t.reset_mock()  # magic call may have been recorded

    rw.start(transaction=mock.sentinel.txn)

    expected = [
        mock.call.send(
            r[1]["queue"],
            downstream_message(1, r["start"][0][1]),
            headers={"workflows-recipe": True},
            transaction=mock.sentinel.txn,
        )
    ]
    assert t.mock_calls == expected


def test_parameter_replacement_in_recipe_wrapper():
    """A RecipeWrapper should support the recipe's .apply_parameters call, and
    this should affect the stored recipe as well as internal referenced parts
    of the recipe."""
    m = generate_recipe_message()

    rw = RecipeWrapper(message=m)

    rw.apply_parameters({"placeholder": "replacement"})

    assert rw.recipe[1]["parameter"] == "some replacement to test replacement"
    assert rw.recipe_step["parameter"] == "some replacement to test replacement"

    assert rw.recipe != Recipe(m["recipe"])
    assert rw.recipe_pointer == m["recipe-pointer"]
    assert rw.recipe_step != m["recipe"][m["recipe-pointer"]]
    assert rw.recipe_path == m["recipe-path"]
    assert rw.environment == m["environment"]
    assert rw.payload == mock.sentinel.payload



import argparse
import importlib
import inspect
import json
import optparse
import os
import tempfile
from unittest import mock

import pytest
import stomp as stomppy
import workflows
import workflows.transport
from workflows.transport.stomp_transport import StompTransport


def test_lookup_and_initialize_stomp_transport_layer():
    """Find the stomp transport layer via the lookup mechanism and run
    its constructor with default settings."""
    stomp = workflows.transport.lookup("StompTransport")
    assert stomp == StompTransport
    stomp()


def test_add_command_line_help_optparse():
    """Check that command line parameters are registered in the parser."""
    parser = mock.MagicMock()

    StompTransport().add_command_line_options(parser)

    parser.add_argument.assert_not_called()
    parser.add_option.assert_called()
    assert parser.add_option.call_count > 4
    for call in parser.add_option.call_args_list:
        assert call[1]["action"] == "callback"


def test_add_command_line_help_argparse():
    """Check that command line parameters are registered in the parser."""
    parser = mock.MagicMock()
    parser.add_argument = mock.Mock()

    StompTransport().add_command_line_options(parser)

    parser.add_argument.assert_called()
    parser.add_option.assert_not_called()
    assert parser.add_argument.call_count > 4
    for call in parser.add_argument.call_args_list:
        assert inspect.isclass(call[1]["action"])


def test_adding_arguments_to_argparser():
    """Check that command line parameters can be added to the parser."""
    parser = argparse.ArgumentParser()

    StompTransport().add_command_line_options(parser)

    result = parser.parse_args([])
    assert result.stomp_host
    assert result.stomp_port
    assert result.stomp_user
    assert result.stomp_pass


@mock.patch("workflows.transport.stomp_transport.stomp")
def test_check_config_file_behaviour(mockstomp):
    """Check that a specified configuration file is read, that command line
    parameters have precedence and are passed on to the stomp layer."""
    mockconn = mock.Mock()
    mockstomp.Connection.return_value = mockconn
    parser = optparse.OptionParser()
    stomp = StompTransport()
    stomp.add_command_line_options(parser)

    # Temporarily create an example stomp configuration file
    cfgfile = tempfile.NamedTemporaryFile(delete=False)
    try:
        cfgfile.write(
            """
# An example stomp configuration file
# Only lines in the [stomp] block will be interpreted

[stomp]
#host = 127.0.0.1
port = 1234
username = someuser
password = somesecret
prefix = namespace
""".encode(
                "utf-8"
            )
        )
        cfgfile.close()

        parser.parse_args(
            ["--stomp-conf", cfgfile.name, "--stomp-user", mock.sentinel.user]
        )

        # Command line parameters are shared for all instances
        stomp = StompTransport()
        stomp.connect()

        # Reset configuration for subsequent tests by reloading StompTransport
        importlib.reload(workflows.transport.stomp_transport)
        globals()["StompTransport"] = workflows.transport.stomp_transport.StompTransport

        mockstomp.Connection.assert_called_once_with([("localhost", 1234)])
        mockconn.connect.assert_called_once_with(
            mock.sentinel.user, "somesecret", wait=False
        )
        assert stomp.get_namespace() == "namespace"

    finally:
        os.remove(cfgfile.name)

    # Loading a non-existing configuration file
    with pytest.raises(workflows.Error):
        parser.parse_args(["--stomp-conf", ""])


@mock.patch("workflows.transport.stomp_transport.stomp")
def test_anonymous_connection(mockstomp):
    """Check that a specified configuration file is read, that command line
    parameters have precedence and are passed on to the stomp layer."""
    mockconn = mock.Mock()
    mockstomp.Connection.return_value = mockconn
    parser = optparse.OptionParser()
    stomp = StompTransport()
    stomp.add_command_line_options(parser)

    parser.parse_args(["--stomp-user=", "--stomp-pass="])

    # Command line parameters are shared for all instances
    stomp = StompTransport()
    stomp.connect()

    # Reset configuration for subsequent tests by reloading StompTransport
    importlib.reload(workflows.transport.stomp_transport)
    globals()["StompTransport"] = workflows.transport.stomp_transport.StompTransport

    mockconn.connect.assert_called_once_with(wait=False)


@mock.patch("workflows.transport.stomp_transport.stomp")
def test_instantiate_link_and_connect_to_broker(mockstomp):
    """Test the Stomp connection routine."""
    stomp = StompTransport()
    mockconn = mockstomp.Connection.return_value

    assert not stomp.is_connected()

    stomp.connect()

    mockstomp.Connection.assert_called_once()
    mockconn.connect.assert_called_once()
    assert stomp.is_connected()

    stomp.connect()

    mockstomp.Connection.assert_called_once()
    mockconn.connect.assert_called_once()
    assert stomp.is_connected()

    stomp.disconnect()

    mockstomp.Connection.assert_called_once()
    mockconn.connect.assert_called_once()
    mockconn.disconnect.assert_called_once()
    assert not stomp.is_connected()

    stomp.disconnect()

    mockstomp.Connection.assert_called_once()
    mockconn.connect.assert_called_once()
    mockconn.disconnect.assert_called_once()
    assert not stomp.is_connected()


@mock.patch("workflows.transport.stomp_transport.stomp")
def test_error_handling_when_connecting_to_broker(mockstomp):
    """Test the Stomp connection routine."""
    stomp = StompTransport()
    mockconn = mockstomp.Connection.return_value
    mockconn.connect.side_effect = stomppy.exception.ConnectFailedException()
    mockstomp.exception.ConnectFailedException = (
        stomppy.exception.ConnectFailedException
    )

    with pytest.raises(workflows.Disconnected):
        stomp.connect()

    assert not stomp.is_connected()


@mock.patch("workflows.transport.stomp_transport.time")
@mock.patch("workflows.transport.stomp_transport.stomp")
def test_broadcast_status(mockstomp, mocktime):
    """Test the status broadcast function."""
    mocktime.time.return_value = 20000
    stomp = StompTransport()
    stomp.connect()
    mockconn = mockstomp.Connection.return_value

    stomp.broadcast_status({"status": str(mock.sentinel.status)})

    mockconn.send.assert_called_once()
    args, kwargs = mockconn.send.call_args
    # expiration should be 15 seconds in the future
    assert int(kwargs["headers"]["expires"]) == 1000 * (20000 + 15)
    destination, message = args
    assert destination.startswith("/topic/transient.status")
    statusdict = json.loads(message)
    assert statusdict["status"] == str(mock.sentinel.status)


@mock.patch("workflows.transport.stomp_transport.stomp")
def test_send_message(mockstomp):
    """Test the message sending function."""
    stomp = StompTransport()
    stomp.connect()
    mockconn = mockstomp.Connection.return_value

    stomp._send(str(mock.sentinel.channel), mock.sentinel.message)

    mockconn.send.assert_called_once()
    args, kwargs = mockconn.send.call_args
    assert args == ("/queue/" + str(mock.sentinel.channel), mock.sentinel.message)
    assert kwargs.get("headers") == {"persistent": "true"}

    stomp._send(
        str(mock.sentinel.channel),
        mock.sentinel.message,
        headers={"hdr": mock.sentinel.header},
        delay=123,
    )
    assert mockconn.send.call_count == 2
    args, kwargs = mockconn.send.call_args
    assert args == ("/queue/" + str(mock.sentinel.channel), mock.sentinel.message)
    assert kwargs == {
        "headers": {
            "hdr": mock.sentinel.header,
            "persistent": "true",
            "AMQ_SCHEDULED_DELAY": 123000,
        }
    }


@mock.patch("workflows.transport.stomp_transport.stomp")
@mock.patch("workflows.transport.stomp_transport.time")
def test_sending_message_with_expiration(time, mockstomp):
    """Test sending a message that expires some time in the future."""
    system_time = 1234567.1234567
    message_lifetime = 120
    expiration_time = int((system_time + message_lifetime) * 1000)
    time.time.return_value = system_time

    stomp = StompTransport()
    stomp.connect()
    mockconn = mockstomp.Connection.return_value

    stomp._send(str(mock.sentinel.channel), mock.sentinel.message, expiration=120)

    mockconn.send.assert_called_once()
    args, kwargs = mockconn.send.call_args
    assert args == ("/queue/" + str(mock.sentinel.channel), mock.sentinel.message)
    assert kwargs.get("headers") == {"persistent": "true", "expires": expiration_time}


@mock.patch("workflows.transport.stomp_transport.stomp")
def test_error_handling_on_send(mockstomp):
    """Unrecoverable errors during sending should mark the connection as disconnected."""
    stomp = StompTransport()
    stomp.connect()
    mockconn = mockstomp.Connection.return_value
    mockconn.send.side_effect = stomppy.exception.NotConnectedException()
    mockstomp.exception = stomppy.exception

    with pytest.raises(workflows.Disconnected):
        stomp._send(str(mock.sentinel.channel), mock.sentinel.message)
    assert not stomp.is_connected()


@mock.patch("workflows.transport.stomp_transport.stomp")
def test_send_broadcast(mockstomp):
    """Test the broadcast sending function."""
    stomp = StompTransport()
    stomp.connect()
    mockconn = mockstomp.Connection.return_value

    stomp._broadcast(str(mock.sentinel.channel), mock.sentinel.message)

    mockconn.send.assert_called_once()
    args, kwargs = mockconn.send.call_args
    assert args == ("/topic/" + str(mock.sentinel.channel), mock.sentinel.message)
    assert kwargs.get("headers") in (None, {})

    stomp._broadcast(
        str(mock.sentinel.channel), mock.sentinel.message, headers=mock.sentinel.headers
    )
    assert mockconn.send.call_count == 2
    args, kwargs = mockconn.send.call_args
    assert args == ("/topic/" + str(mock.sentinel.channel), mock.sentinel.message)
    assert kwargs == {"headers": mock.sentinel.headers}

    stomp._broadcast(str(mock.sentinel.channel), mock.sentinel.message, delay=123)
    assert mockconn.send.call_count == 3
    args, kwargs = mockconn.send.call_args
    assert args == ("/topic/" + str(mock.sentinel.channel), mock.sentinel.message)
    assert kwargs["headers"].get("AMQ_SCHEDULED_DELAY") == 123000


@mock.patch("workflows.transport.stomp_transport.stomp")
@mock.patch("workflows.transport.stomp_transport.time")
def test_broadcasting_message_with_expiration(time, mockstomp):
    """Test sending a message that expires some time in the future."""
    system_time = 1234567.1234567
    message_lifetime = 120
    expiration_time = int((system_time + message_lifetime) * 1000)
    time.time.return_value = system_time

    stomp = StompTransport()
    stomp.connect()
    mockconn = mockstomp.Connection.return_value

    stomp._broadcast(str(mock.sentinel.channel), mock.sentinel.message, expiration=120)

    mockconn.send.assert_called_once()
    args, kwargs = mockconn.send.call_args
    assert args == ("/topic/" + str(mock.sentinel.channel), mock.sentinel.message)
    assert kwargs.get("headers") == {"expires": expiration_time}


@mock.patch("workflows.transport.stomp_transport.stomp")
def test_error_handling_on_broadcast(mockstomp):
    """Unrecoverable errors during broadcasting should mark the connection as disconnected."""
    stomp = StompTransport()
    stomp.connect()
    mockconn = mockstomp.Connection.return_value
    mockconn.send.side_effect = stomppy.exception.NotConnectedException()
    mockstomp.exception = stomppy.exception

    with pytest.raises(stomppy.exception.NotConnectedException):
        stomp._broadcast(str(mock.sentinel.channel), mock.sentinel.message)
    assert not stomp.is_connected()


@mock.patch("workflows.transport.stomp_transport.stomp")
def test_messages_are_serialized_for_transport(mockstomp):
    """Test the message serialization."""
    banana = {"entry": [0, "banana"]}
    banana_str = '{"entry": [0, "banana"]}'
    stomp = StompTransport()
    stomp.connect()
    mockconn = mockstomp.Connection.return_value

    stomp.send(str(mock.sentinel.channel1), banana)
    mockconn.send.assert_called_once()
    args, kwargs = mockconn.send.call_args
    assert args == ("/queue/" + str(mock.sentinel.channel1), banana_str)

    stomp.broadcast(str(mock.sentinel.channel2), banana)
    args, kwargs = mockconn.send.call_args
    assert args == ("/topic/" + str(mock.sentinel.channel2), banana_str)

    with pytest.raises(Exception):
        stomp.send(str(mock.sentinel.channel), mock.sentinel.unserializable)


@mock.patch("workflows.transport.stomp_transport.stomp")
def test_messages_are_not_serialized_for_raw_transport(mockstomp):
    """Test the raw sending methods."""
    banana = '{"entry": [0, "banana"]}'
    stomp = StompTransport()
    stomp.connect()
    mockconn = mockstomp.Connection.return_value

    stomp.raw_send(str(mock.sentinel.channel1), banana)
    mockconn.send.assert_called_once()
    args, kwargs = mockconn.send.call_args
    assert args == ("/queue/" + str(mock.sentinel.channel1), banana)

    mockconn.send.reset_mock()
    stomp.raw_broadcast(str(mock.sentinel.channel2), banana)
    mockconn.send.assert_called_once()
    args, kwargs = mockconn.send.call_args
    assert args == ("/topic/" + str(mock.sentinel.channel2), banana)

    mockconn.send.reset_mock()
    stomp.raw_send(str(mock.sentinel.channel), mock.sentinel.unserializable)
    mockconn.send.assert_called_once()
    args, kwargs = mockconn.send.call_args
    assert args == (
        "/queue/" + str(mock.sentinel.channel),
        mock.sentinel.unserializable,
    )


@mock.patch("workflows.transport.stomp_transport.stomp")
def test_messages_are_deserialized_after_transport(mockstomp):
    """Test the message serialization."""
    banana = {"entry": [0, "banana"]}
    banana_str = '{"entry": [0, "banana"]}'
    stomp = StompTransport()
    stomp.connect()
    mockconn = mockstomp.Connection.return_value
    message_handler = mockconn.set_listener.call_args[0][1].on_message

    # Test subscriptions
    callback = mock.Mock()
    stomp.subscribe("channel", callback)
    subscription_id = mockconn.subscribe.call_args[0][1]
    message_handler({"subscription": subscription_id}, banana_str)
    callback.assert_called_once_with({"subscription": subscription_id}, banana)

    message_handler({"subscription": subscription_id}, mock.sentinel.undeserializable)
    callback.assert_called_with(
        {"subscription": subscription_id}, mock.sentinel.undeserializable
    )

    # Test broadcast subscriptions
    callback = mock.Mock()
    stomp.subscribe_broadcast("channel", callback)
    subscription_id = mockconn.subscribe.call_args[0][1]
    message_handler({"subscription": subscription_id}, banana_str)
    callback.assert_called_once_with({"subscription": subscription_id}, banana)

    message_handler({"subscription": subscription_id}, mock.sentinel.undeserializable)
    callback.assert_called_with(
        {"subscription": subscription_id}, mock.sentinel.undeserializable
    )

    # Test subscriptions with mangling disabled
    callback = mock.Mock()
    stomp.subscribe("channel", callback, disable_mangling=True)
    subscription_id = mockconn.subscribe.call_args[0][1]
    message_handler({"subscription": subscription_id}, banana_str)
    callback.assert_called_once_with({"subscription": subscription_id}, banana_str)

    # Test broadcast subscriptions with mangling disabled
    callback = mock.Mock()
    stomp.subscribe_broadcast("channel", callback, disable_mangling=True)
    subscription_id = mockconn.subscribe.call_args[0][1]
    message_handler({"subscription": subscription_id}, banana_str)
    callback.assert_called_once_with({"subscription": subscription_id}, banana_str)


@mock.patch("workflows.transport.stomp_transport.stomp")
def test_subscribe_to_queue(mockstomp):
    """Test subscribing to a queue (producer-consumer), callback functions and unsubscribe."""
    mock_cb1 = mock.Mock()
    mock_cb2 = mock.Mock()
    stomp = StompTransport()
    stomp.connect()
    mockconn = mockstomp.Connection.return_value

    def callback_resolver(cbid):
        if cbid == 1:
            return mock_cb1
        if cbid == 2:
            return mock_cb2
        raise ValueError("Unknown subscription ID %r" % cbid)

    stomp.subscription_callback = callback_resolver

    mockconn.set_listener.assert_called_once()
    listener = mockconn.set_listener.call_args[0][1]
    assert listener is not None

    stomp._subscribe(
        1,
        str(mock.sentinel.channel1),
        mock_cb1,
        transformation=mock.sentinel.transformation,
    )

    mockconn.subscribe.assert_called_once()
    args, kwargs = mockconn.subscribe.call_args
    assert args == ("/queue/" + str(mock.sentinel.channel1), 1)
    assert kwargs == {
        "headers": {"transformation": mock.sentinel.transformation},
        "ack": "auto",
    }

    stomp._subscribe(
        2,
        str(mock.sentinel.channel2),
        mock_cb2,
        retroactive=True,
        selector=mock.sentinel.selector,
        exclusive=True,
        transformation=True,
        priority=42,
    )
    assert mockconn.subscribe.call_count == 2
    args, kwargs = mockconn.subscribe.call_args
    assert args == ("/queue/" + str(mock.sentinel.channel2), 2)
    assert kwargs == {
        "headers": {
            "activemq.retroactive": "true",
            "selector": mock.sentinel.selector,
            "activemq.exclusive": "true",
            "transformation": "jms-object-json",
            "activemq.priority": 42,
        },
        "ack": "auto",
    }

    assert mock_cb1.call_count == 0
    listener.on_message({"subscription": 1}, mock.sentinel.message1)
    mock_cb1.assert_called_once_with({"subscription": 1}, mock.sentinel.message1)

    assert mock_cb2.call_count == 0
    listener.on_message({"subscription": 2}, mock.sentinel.message2)
    mock_cb2.assert_called_once_with({"subscription": 2}, mock.sentinel.message2)

    stomp._subscribe(3, str(mock.sentinel.channel3), mock_cb2, acknowledgement=True)
    assert mockconn.subscribe.call_count == 3
    args, kwargs = mockconn.subscribe.call_args
    assert args == ("/queue/" + str(mock.sentinel.channel3), 3)
    assert kwargs == {"headers": {}, "ack": "client-individual"}

    stomp._unsubscribe(1)
    mockconn.unsubscribe.assert_called_once_with(id=1)
    stomp._unsubscribe(2)
    mockconn.unsubscribe.assert_called_with(id=2)


@mock.patch("workflows.transport.stomp_transport.stomp")
def test_subscribe_to_broadcast(mockstomp):
    """Test subscribing to a topic (publish-subscribe) and callback functions."""
    mock_cb1 = mock.Mock()
    mock_cb2 = mock.Mock()
    stomp = StompTransport()
    stomp.connect()
    mockconn = mockstomp.Connection.return_value

    def callback_resolver(cbid):
        if cbid == 1:
            return mock_cb1
        if cbid == 2:
            return mock_cb2
        raise ValueError("Unknown subscription ID %r" % cbid)

    stomp.subscription_callback = callback_resolver

    mockconn.set_listener.assert_called_once()
    listener = mockconn.set_listener.call_args[0][1]
    assert listener is not None

    stomp._subscribe_broadcast(
        1,
        str(mock.sentinel.channel1),
        mock_cb1,
        transformation=mock.sentinel.transformation,
    )

    mockconn.subscribe.assert_called_once()
    args, kwargs = mockconn.subscribe.call_args
    assert args == ("/topic/" + str(mock.sentinel.channel1), 1)
    assert kwargs == {"headers": {"transformation": mock.sentinel.transformation}}

    stomp._subscribe_broadcast(
        2, str(mock.sentinel.channel2), mock_cb2, retroactive=True, transformation=True
    )
    assert mockconn.subscribe.call_count == 2
    args, kwargs = mockconn.subscribe.call_args
    assert args == ("/topic/" + str(mock.sentinel.channel2), 2)
    assert kwargs == {
        "headers": {"activemq.retroactive": "true", "transformation": "jms-object-json"}
    }

    assert mock_cb1.call_count == 0
    listener.on_message({"subscription": 1}, mock.sentinel.message1)
    mock_cb1.assert_called_once_with({"subscription": 1}, mock.sentinel.message1)

    assert mock_cb2.call_count == 0
    listener.on_message({"subscription": 2}, mock.sentinel.message2)
    mock_cb2.assert_called_once_with({"subscription": 2}, mock.sentinel.message2)

    stomp._unsubscribe(1)
    mockconn.unsubscribe.assert_called_once_with(id=1)
    stomp._unsubscribe(2)
    mockconn.unsubscribe.assert_called_with(id=2)


@mock.patch("workflows.transport.stomp_transport.stomp")
def test_transaction_calls(mockstomp):
    """Test that calls to create, commit, abort transactions are passed to stomp properly."""
    stomp = StompTransport()
    stomp.connect()
    mockconn = mockstomp.Connection.return_value

    stomp._transaction_begin(mock.sentinel.txid)
    mockconn.begin.assert_called_once_with(transaction=mock.sentinel.txid)

    stomp._send("destination", mock.sentinel.message, transaction=mock.sentinel.txid)
    mockconn.send.assert_called_once_with(
        "/queue/destination",
        mock.sentinel.message,
        headers={"persistent": "true"},
        transaction=mock.sentinel.txid,
    )

    stomp._transaction_abort(mock.sentinel.txid)
    mockconn.abort.assert_called_once_with(mock.sentinel.txid)

    stomp._transaction_commit(mock.sentinel.txid)
    mockconn.commit.assert_called_once_with(mock.sentinel.txid)


@mock.patch("workflows.transport.stomp_transport.stomp")
def test_ack_message(mockstomp):
    """Test that the _ack function is properly forwarded to stomp."""
    stomp = StompTransport()
    stomp.connect()
    mockconn = mockstomp.Connection.return_value

    subid = stomp._subscribe(1, str(mock.sentinel.channel3), None, acknowledgement=True)
    stomp._ack(mock.sentinel.messageid, subid)
    mockconn.ack.assert_called_once_with(mock.sentinel.messageid, subid)

    stomp._ack(mock.sentinel.messageid, subid, transaction=mock.sentinel.txn)
    mockconn.ack.assert_called_with(
        mock.sentinel.messageid, subid, transaction=mock.sentinel.txn
    )


@mock.patch("workflows.transport.stomp_transport.stomp")
def test_nack_message(mockstomp):
    """Test that the _nack function is properly forwarded to stomp."""
    stomp = StompTransport()
    stomp.connect()
    mockconn = mockstomp.Connection.return_value

    subid = stomp._subscribe(1, str(mock.sentinel.channel3), None, acknowledgement=True)
    stomp._nack(mock.sentinel.messageid, subid)
    mockconn.nack.assert_called_once_with(mock.sentinel.messageid, subid)

    stomp._nack(mock.sentinel.messageid, subid, transaction=mock.sentinel.txn)
    mockconn.nack.assert_called_with(
        mock.sentinel.messageid, subid, transaction=mock.sentinel.txn
    )


@mock.patch("workflows.transport.stomp_transport.stomp")
def test_namespace_is_used_correctly(mockstomp):
    """Test that a configured namespace is correctly used when subscribing and sending messages."""
    mockconn = mockstomp.Connection.return_value
    StompTransport.defaults["--stomp-prfx"] = ""
    stomp = StompTransport()
    stomp.connect()
    assert stomp.get_namespace() == ""

    StompTransport.defaults["--stomp-prfx"] = "ns."
    stomp = StompTransport()
    stomp.connect()
    assert stomp.get_namespace() == "ns"

    stomp._send("some_queue", mock.sentinel.message1)
    mockconn.send.assert_called_once()
    assert mockconn.send.call_args[0] == (
        "/queue/ns.some_queue",
        mock.sentinel.message1,
    )

    stomp._send("some_queue", mock.sentinel.message2, ignore_namespace=True)
    assert mockconn.send.call_args[0] == ("/queue/some_queue", mock.sentinel.message2)

    StompTransport.defaults["--stomp-prfx"] = "ns"
    stomp = StompTransport()
    stomp.connect()
    assert stomp.get_namespace() == "ns"

    stomp._send("some_queue", mock.sentinel.message1)
    assert mockconn.send.call_args[0] == (
        "/queue/ns.some_queue",
        mock.sentinel.message1,
    )

    stomp._broadcast("some_topic", mock.sentinel.message2)
    assert mockconn.send.call_args[0] == (
        "/topic/ns.some_topic",
        mock.sentinel.message2,
    )

    stomp._broadcast("some_topic", mock.sentinel.message3, ignore_namespace=True)
    assert mockconn.send.call_args[0] == ("/topic/some_topic", mock.sentinel.message3)

    stomp._subscribe(1, "sub_queue", None)
    mockconn.subscribe.assert_called_once()
    assert mockconn.subscribe.call_args[0] == ("/queue/ns.sub_queue", 1)

    stomp._subscribe(2, "sub_queue", None, ignore_namespace=True)
    assert mockconn.subscribe.call_args[0] == ("/queue/sub_queue", 2)

    stomp._subscribe_broadcast(3, "sub_topic", None)
    assert mockconn.subscribe.call_args[0] == ("/topic/ns.sub_topic", 3)

    stomp._subscribe_broadcast(4, "sub_topic", None, ignore_namespace=True)
    assert mockconn.subscribe.call_args[0] == ("/topic/sub_topic", 4)

    stomp.broadcast_status("some status")
    assert mockconn.send.call_args[0] == ("/topic/ns.transient.status", '"some status"')



import time

from workflows.services.common_service import CommonService


class UtilizationStatistics:
    """Generate statistics about the percentage of time spent in different
    statuses over a fixed time slice. This class is not thread-safe."""

    def __init__(self, summation_period=10):
        """Reports will always cover the most recent period of summation_period
        seconds."""
        self.period = summation_period
        self.status_history = [
            {"start": 0, "end": None, "status": CommonService.SERVICE_STATUS_NEW}
        ]

    def update_status(self, new_status):
        """Record a status change with a current timestamp."""
        timestamp = time.time()
        self.status_history[-1]["end"] = timestamp
        self.status_history.append(
            {"start": timestamp, "end": None, "status": new_status}
        )

    def report(self):
        """Return a dictionary of different status codes and the percentage of time
        spent in each throughout the last summation_period seconds.
        Truncate the aggregated history appropriately."""
        timestamp = time.time()
        cutoff = timestamp - self.period
        truncate = 0
        summary = {}
        for event in self.status_history[:-1]:
            if event["end"] < cutoff:
                truncate = truncate + 1
                continue
            summary[event["status"]] = (
                summary.get(event["status"], 0)
                + event["end"]
                - max(cutoff, event["start"])
            )
        summary[self.status_history[-1]["status"]] = (
            summary.get(self.status_history[-1]["status"], 0)
            + timestamp
            - max(cutoff, self.status_history[-1]["start"])
        )
        if truncate:
            self.status_history = self.status_history[truncate:]
        total_duration = sum(summary.values())
        summary = {s: round(d / total_duration, 4) for s, d in summary.items()}
        return summary



import json
import time

from workflows.services.common_service import CommonService


class SampleConsumer(CommonService):
    """An example service building on top of the workflow.services architecture,
    demonstrating how this architecture can be used.
    This service consumes messages off a queue."""

    # Human readable service name
    _service_name = "Message Consumer"

    # Logger name
    _logger_name = "workflows.service.sample_consumer"

    def initializing(self):
        """Subscribe to a channel."""
        self._transport.subscribe("transient.destination", self.consume_message)

    def consume_message(self, header, message):
        """Consume a message"""
        logmessage = {
            "time": (time.time() % 1000) * 1000,
            "header": "",
            "message": message,
        }
        if header:
            logmessage["header"] = (
                json.dumps(header, indent=2) + "\n" + "----------------" + "\n"
            )
        if isinstance(message, dict):
            logmessage["message"] = (
                json.dumps(message, indent=2) + "\n" + "----------------" + "\n"
            )

        print("=== Consume ====\n{header}{message}".format(**logmessage))
        self.log.info("Received message @{time}".format(**logmessage))
        self.log.debug(
            "Received message @{time}\n{header}{message}".format(**logmessage)
        )
        time.sleep(0.1)



"""
Tests the functionality of validate.py with several different recipes which should raise different errors
"""

import json
import sys
from unittest import mock

import pytest
import workflows
from workflows.recipe.validate import validate_recipe, main


def test_validate_returns_type_error_when_called_without_parameters():
    with pytest.raises(TypeError):
        validate_recipe()


def test_no_errors_when_validating_healthy_recipe(tmpdir):
    healthy_recipe = """
        {
          "1": {
            "queue": "simpleservice.submission",
            "parameters": {
              "commands": [
                "echo This is a command32"
              ],
              "workingdir": "/dls/tmp/riw56156/zocalo",
              "output_file": "out.txt"
            }
          },
          "start": [
            [1, []]
          ]
        }
    """
    recipe_file = tmpdir.join("recipe.json")
    recipe_file.write(healthy_recipe)

    validate_recipe(recipe_file.strpath)


def test_value_error_when_validating_bad_json(tmpdir):
    bad_json = """
        {
          "1": {
            "queue": "simpleservice.submission"
            "parameters": {
              "commands": [
                "echo This is a command32"
              ],
              "workingdir": "/dls/tmp/riw56156/zocalo",
              "output_file": "out.txt"
            }
          },
          "start": [
            [1, []]
          ]
        }
    """
    recipe_file = tmpdir.join("recipe.json")
    recipe_file.write(bad_json)

    # Run validate with mock open, expect JSON error
    with pytest.raises(json.JSONDecodeError):
        validate_recipe(recipe_file.strpath)


def test_workflows_error_when_validating_incorrect_workflows_recipe(tmpdir):
    bad_recipe = """
        {
          "1": {
            "queue": "simpleservice.submission",
            "parameters": {
              "commands": [
                "echo This is a command32"
              ],
              "workingdir": "/dls/tmp/riw56156/zocalo",
              "output_file": "out.txt"
            }
          }
        }
    """
    recipe_file = tmpdir.join("recipe.json")
    recipe_file.write(bad_recipe)

    # Run validate with mock open, expect JSON error
    with pytest.raises(workflows.Error):
        validate_recipe(recipe_file.strpath)


# Create a mock of the validate call
@mock.patch("workflows.recipe.validate.validate_recipe")
def test_command_line_validation_one_argument(mock_requests):
    # Create fake arguments to test on
    test_args = ["validate.py", "file1"]
    with mock.patch.object(sys, "argv", test_args):
        main()
    mock_requests.assert_called_once_with("file1")


def test_exit_when_command_line_validation_given_no_arguments():
    # Create fake arguments to test on
    test_args = ["validate.py"]
    with pytest.raises(SystemExit):
        with mock.patch.object(sys, "argv", test_args):
            main()


# Create a mock of the validate call
@mock.patch("workflows.recipe.validate.validate_recipe")
def test_command_line_validation_multiple_arguments(mock_requests):
    # Create fake arguments to test on
    test_args = ["validate.py", "file1", "file2", "file3"]
    with mock.patch.object(sys, "argv", test_args):
        main()
    mock_requests.assert_has_calls(
        [mock.call("file1"), mock.call("file2"), mock.call("file3")]
    )
    assert mock_requests.call_count == 3


@mock.patch("workflows.recipe.validate.validate_recipe")
def test_system_exit_when_error_raised_from_command_line_validation(mock_requests):
    mock_requests.get.side_effect = Exception
    # Create fake arguments to test on
    test_args = ["validate.py"]
    with pytest.raises(SystemExit):
        with mock.patch.object(sys, "argv", test_args):
            main()



import pytest
import workflows.frontend
from unittest import mock
from workflows.services.common_service import CommonService

### Helper classes used in tests ##############################################


class ServiceCrashingOnInit(CommonService):
    """A service that raises an unhandled exception."""

    @staticmethod
    def initializing():
        """Raise AssertionError.
        This should set the error state, kill the service and cause the frontend
        to leave its main loop."""
        assert False  # pragma: no cover


class MockPipe:
    """An object that behaves like a pipe."""

    def __init__(self, contents):
        """Load up contents into pipe."""
        self.contents = contents

    @staticmethod
    def poll(time=None):
        """Check if pipe is empty. There is always something there, either
        some content or an EOFError."""
        return True

    def recv(self):
        """Return first item off the list or raise exception.
        Call callback function if defined and pipe is emptied."""
        if not self.contents:
            raise EOFError("Pipe is empty")
        return_value = self.contents.pop(0)
        if isinstance(return_value, Exception):
            raise return_value
        return return_value

    @staticmethod
    def close():
        """This pipe can't be written to anyway. Ignore call."""

    def assert_empty(self):
        """Pipe must not contain any valid messages. Exceptions may remain."""
        assert all(isinstance(m, Exception) for m in self.contents)


def assert_single_call_only(target):
    def wrapper(*args, **kwargs):
        if hasattr(wrapper, "called"):
            raise Exception("Only a single call to object %s is allowed" % str(target))
        setattr(wrapper, "called", True)
        return target(*args, **kwargs)

    return wrapper


###############################################################################


@mock.patch("workflows.frontend.multiprocessing")
@mock.patch("workflows.frontend.workflows.transport")
def test_frontend_connects_to_transport_layer(mock_transport, mock_mp):
    """Frontend should call connect method on transport layer module and subscribe to a unique command queue."""
    workflows.frontend.Frontend()
    mock_transport.lookup.assert_called_once_with(None)
    mock_transport.lookup.return_value.assert_called_once_with()
    mock_transport.lookup.return_value.return_value.connect.assert_called_once_with()

    workflows.frontend.Frontend(transport="some_transport")
    mock_transport.lookup.assert_called_with("some_transport")

    mock_trn = mock.Mock()
    workflows.frontend.Frontend(mock_trn)
    mock_trn.assert_called_once_with()
    mock_trn.return_value.connect.assert_called_once_with()


@mock.patch("workflows.frontend.workflows.transport")
def test_frontend_subscribes_to_command_channel(mock_transport):
    """Frontend should call connect method on transport layer module and subscribe to a unique command queue."""
    transport = mock_transport.lookup.return_value.return_value
    fe = workflows.frontend.Frontend()
    mock_transport.lookup.assert_called_once_with(None)
    transport.connect.assert_called_once()
    transport.subscribe.assert_not_called()

    def frontend_with_message(message):
        transport.reset_mock()
        fe = workflows.frontend.Frontend(
            transport_command_channel=mock.sentinel.command
        )
        transport.subscribe_broadcast.assert_called_once_with(
            mock.sentinel.command, mock.ANY
        )
        transport.subscribe_broadcast.call_args[0][1]({}, message)
        return fe

    assert frontend_with_message({}).shutdown is False
    assert frontend_with_message({"command": "shutdown"}).shutdown is False
    assert (
        frontend_with_message(
            {"command": "shutdown", "host": fe.get_host_id()}
        ).shutdown
        is True
    )
    assert (
        frontend_with_message(
            {"command": "shutdown", "host": mock.sentinel.different_host}
        ).shutdown
        is False
    )
    assert (
        frontend_with_message(
            {"command": "shutdown", "service": fe.get_status()["serviceclass"]}
        ).shutdown
        is True
    )
    assert (
        frontend_with_message(
            {"command": "shutdown", "service": mock.sentinel.different_service}
        ).shutdown
        is False
    )


@mock.patch("workflows.frontend.multiprocessing")
@mock.patch("workflows.frontend.workflows.transport")
def test_start_service_in_frontend(mock_transport, mock_mp):
    """Check that the service is being run and connected to the frontend via the correct pipes."""
    mock_service = mock.Mock()
    pipes = [mock.Mock(), mock.Mock(), mock.Mock(), mock.Mock()]
    mock_mp.Pipe.side_effect = [(pipes[0], pipes[1]), (pipes[2], pipes[3]), None]

    # initialize frontend
    fe = workflows.frontend.Frontend(environment=mock.sentinel.environment)

    # start service
    fe.switch_service(mock_service)

    # check service was started properly
    mock_service.assert_called_once_with(environment=mock.sentinel.environment)
    mock_service.return_value.connect.assert_called_once_with(
        commands=pipes[0], frontend=pipes[3]
    )
    mock_mp.Process.assert_called_once_with(
        target=mock_service.return_value.start, args=(), kwargs=mock.ANY
    )
    mock_mp.Process.return_value.start.assert_called_once()

    # Fun with python multiprocessing:
    # Because the pipe file descriptors are copied they must be closed in the process not using them
    pipes[0].close.assert_called_once_with()
    pipes[3].close.assert_called_once_with()


@mock.patch("workflows.frontend.workflows.transport")
def test_get_frontend_status(mock_transport):
    """Check that the get_status-method works and contains the correct host-id."""
    fe = workflows.frontend.Frontend()
    status = fe.get_status()
    assert status["host"] == fe.get_host_id()


@pytest.mark.timeout(3)
def test_frontend_can_handle_unhandled_service_initialization_exceptions():
    """When a service crashes on initialization an exception should be thrown."""
    transport = mock.Mock()

    fe = workflows.frontend.Frontend(
        transport=transport, service=assert_single_call_only(ServiceCrashingOnInit)
    )
    transport = transport.return_value
    transport.connect.assert_called_once()

    with pytest.raises(workflows.Error):
        fe.run()

    status_list = [
        args[0].get("status")
        for args, kwargs in transport.broadcast_status.call_args_list
        if args
    ]
    assert status_list == [
        CommonService.SERVICE_STATUS_NEW,
        CommonService.SERVICE_STATUS_STARTING,
        CommonService.SERVICE_STATUS_ERROR,
        CommonService.SERVICE_STATUS_END,
    ]


@mock.patch("workflows.frontend.multiprocessing")
def test_frontend_can_handle_service_initialization_segfaults(mock_mp):
    """When a service crashes on initialization an exception should be thrown."""
    transport = mock.Mock()
    service = mock.Mock()
    service_process = mock.Mock()
    dummy_pipe = mock.Mock()
    dummy_pipe.recv.side_effect = EOFError()  # Dead on arrival
    mock_mp.Pipe.return_value = (dummy_pipe, dummy_pipe)
    mock_mp.Process.return_value = service_process

    fe = workflows.frontend.Frontend(transport=transport, service=service)
    transport = transport.return_value
    transport.connect.assert_called_once()
    mock_mp.Process.assert_called_once_with(
        target=service.return_value.start, args=(), kwargs=mock.ANY
    )

    with pytest.raises(workflows.Error):
        fe.run()
    service_process.start.assert_called()
    service_process.join.assert_called()


@mock.patch("workflows.frontend.multiprocessing")
def test_frontend_terminates_on_transport_disconnection(mock_mp):
    """When the transport connection is lost permanently, the frontend should stop."""
    transport = mock.Mock()
    service = mock.Mock()
    service_process = mock.Mock()
    dummy_pipe = mock.Mock()
    dummy_pipe.poll.side_effect = [
        False,
        Exception("Frontend did not terminate on transport disconnect"),
    ]
    mock_mp.Pipe.return_value = (dummy_pipe, dummy_pipe)
    mock_mp.Process.return_value = service_process
    service_process.is_alive.return_value = True

    fe = workflows.frontend.Frontend(transport=transport, service=service)
    transport = transport.return_value
    transport.connect.assert_called_once()
    transport.is_connected.return_value = False
    mock_mp.Process.assert_called_once_with(
        target=service.return_value.start, args=(), kwargs=mock.ANY
    )

    with pytest.raises(workflows.Error):
        fe.run()

    service_process.terminate.assert_called()
    service_process.join.assert_called()


@mock.patch("workflows.frontend.multiprocessing")
def test_frontend_parses_status_updates(mock_mp):
    """The frontend should forward status updates to the advertiser thread when appropriate."""
    transport = mock.Mock()
    service_process = mock.Mock()
    mock_mp.Process.return_value = service_process

    dummy_pipe = MockPipe(
        [
            {"band": "status_update", "statuscode": CommonService.SERVICE_STATUS_NEW},
            {
                "band": "status_update",
                "statuscode": CommonService.SERVICE_STATUS_STARTING,
            },
            {
                "band": "status_update",
                "statuscode": CommonService.SERVICE_STATUS_PROCESSING,
            },
            {
                "band": "status_update",
                "statuscode": CommonService.SERVICE_STATUS_SHUTDOWN,
            },
            {"band": "status_update", "statuscode": CommonService.SERVICE_STATUS_END},
        ]
    )
    mock_mp.Pipe.return_value = (dummy_pipe, dummy_pipe)

    fe = workflows.frontend.Frontend(transport=transport, service=mock.Mock())

    # intercept status code updates
    status_list = []
    original_status_update_fn = fe.update_status

    def intercept(*args, **kwargs):
        if "status_code" in kwargs:
            status_list.append(kwargs["status_code"])
        return original_status_update_fn(*args, **kwargs)

    fe.update_status = intercept

    fe.run()

    dummy_pipe.assert_empty()
    assert status_list == [
        CommonService.SERVICE_STATUS_NEW,
        CommonService.SERVICE_STATUS_STARTING,
        CommonService.SERVICE_STATUS_PROCESSING,
        CommonService.SERVICE_STATUS_SHUTDOWN,
        CommonService.SERVICE_STATUS_END,  # Following updates caused by frontend
        CommonService.SERVICE_STATUS_END,
        CommonService.SERVICE_STATUS_TEARDOWN,
    ]


@mock.patch("workflows.frontend.multiprocessing")
def test_frontend_parses_termination_requests(mock_mp):
    """The frontend should observe termination requests from the server."""
    transport = mock.Mock()
    service_process = mock.Mock()
    mock_mp.Process.return_value = service_process

    dummy_pipe = MockPipe(
        [
            {"band": "status_update", "statuscode": CommonService.SERVICE_STATUS_NEW},
            {
                "band": "status_update",
                "statuscode": CommonService.SERVICE_STATUS_STARTING,
            },
            {
                "band": "status_update",
                "statuscode": CommonService.SERVICE_STATUS_PROCESSING,
            },
            {"band": "request_termination"},
            RuntimeError("Service termination request not observed"),
        ]
    )
    mock_mp.Pipe.return_value = (dummy_pipe, dummy_pipe)

    fe = workflows.frontend.Frontend(transport=transport, service=mock.Mock())

    # intercept status code updates
    status_list = []
    original_status_update_fn = fe.update_status

    def intercept(*args, **kwargs):
        if "status_code" in kwargs:
            status_list.append(kwargs["status_code"])
        return original_status_update_fn(*args, **kwargs)

    fe.update_status = intercept

    fe.run()

    dummy_pipe.assert_empty()
    assert status_list == [
        CommonService.SERVICE_STATUS_NEW,
        CommonService.SERVICE_STATUS_STARTING,
        CommonService.SERVICE_STATUS_PROCESSING,
        CommonService.SERVICE_STATUS_END,
        CommonService.SERVICE_STATUS_TEARDOWN,
    ]


@mock.patch("workflows.frontend.time")
@mock.patch("workflows.frontend.multiprocessing")
def test_frontend_sends_status_updates(mock_mp, mock_time):
    """The frontend should send status updates on update_status() calls.
    Some sensible rate limiting must be applied."""
    transport = mock.Mock()
    mock_time.time.return_value = 10

    fe = workflows.frontend.Frontend(transport=transport)

    transport = transport.return_value
    transport.broadcast_status.assert_called_once()
    assert (
        transport.broadcast_status.call_args[0][0]["status"]
        == CommonService.SERVICE_STATUS_NONE
    )
    transport.broadcast_status.reset_mock()

    fe.update_status(status_code=CommonService.SERVICE_STATUS_STARTING)

    transport.broadcast_status.assert_called_once()
    assert (
        transport.broadcast_status.call_args[0][0]["status"]
        == CommonService.SERVICE_STATUS_STARTING
    )
    transport.broadcast_status.reset_mock()

    # subsequent update call that does not change anything should be ignored
    fe.update_status()
    transport.broadcast_status.assert_not_called()

    # time passes. Update call should cause broadcast.
    mock_time.time.return_value = 20
    fe.update_status()
    transport.broadcast_status.assert_called_once()
    assert (
        transport.broadcast_status.call_args[0][0]["status"]
        == CommonService.SERVICE_STATUS_STARTING
    )
    transport.broadcast_status.reset_mock()

    # not much time passes. Update call should still cause broadcast, because status has not been seen before
    mock_time.time.return_value = 20.1
    fe.update_status(status_code=CommonService.SERVICE_STATUS_PROCESSING)
    transport.broadcast_status.assert_called_once()
    assert (
        transport.broadcast_status.call_args[0][0]["status"]
        == CommonService.SERVICE_STATUS_PROCESSING
    )
    transport.broadcast_status.reset_mock()

    # not much time passes. Update call should not broadcast, because this is a recent IDLE status
    mock_time.time.return_value = 20.2
    fe.update_status(status_code=CommonService.SERVICE_STATUS_IDLE)
    transport.broadcast_status.assert_not_called()

    # not much time passes. Update call should not broadcast, because the announced status has not changed
    mock_time.time.return_value = 20.3
    fe.update_status(status_code=CommonService.SERVICE_STATUS_PROCESSING)
    transport.broadcast_status.assert_not_called()
    transport.broadcast_status.reset_mock()

    # not much time passes. Update call should still not broadcast, because this is again a recent IDLE status
    mock_time.time.return_value = 20.4
    fe.update_status(status_code=CommonService.SERVICE_STATUS_IDLE)
    transport.broadcast_status.assert_not_called()

    # not much time passes. Update call should still not broadcast, because this IDLE is still too recent
    mock_time.time.return_value = 20.8
    fe.update_status(status_code=CommonService.SERVICE_STATUS_IDLE)
    transport.broadcast_status.assert_not_called()

    # however as time passes the update call should cause a late broadcast
    mock_time.time.return_value = 21
    fe.update_status()
    transport.broadcast_status.assert_called_once()
    assert (
        transport.broadcast_status.call_args[0][0]["status"]
        == CommonService.SERVICE_STATUS_IDLE
    )


@mock.patch("workflows.frontend.multiprocessing")
def test_frontend_does_not_restart_nonrestartable_service_on_segfault(mock_mp):
    """When the frontend is constructed with restart_service=False failing services must not be restarted."""
    service_factory = mock.Mock()
    service_process = mock.Mock()
    dummy_pipe = mock.Mock()
    dummy_pipe.recv.side_effect = EOFError()  # Dead on arrival
    mock_mp.Pipe.return_value = (dummy_pipe, dummy_pipe)
    mock_mp.Process.return_value = service_process

    service_instances = [mock.Mock()]
    service_factory.side_effect = service_instances + [
        Exception("More than one service object instantiated")
    ]

    fe = workflows.frontend.Frontend(
        transport=mock.Mock(), service=service_factory, restart_service=False
    )
    with pytest.raises(workflows.Error):
        fe.run()

    service_factory.assert_called_once()
    service_process.start.assert_called_once()
    service_process.join.assert_called_once()
    mock_mp.Process.assert_called_once_with(
        target=service_instances[0].start, args=(), kwargs=mock.ANY
    )


@mock.patch("workflows.frontend.multiprocessing")
def test_frontend_does_not_restart_nonrestartable_service_on_error(mock_mp):
    """When the frontend is constructed with restart_service=False failing services must not be restarted."""
    service_instances = [mock.Mock()]
    service_factory = mock.Mock()
    service_factory.side_effect = service_instances + [
        Exception("More than one service object instantiated")
    ]
    service_process = mock.Mock()
    mock_mp.Process.return_value = service_process

    dummy_pipe = MockPipe(
        [
            {"band": "status_update", "statuscode": CommonService.SERVICE_STATUS_NEW},
            {
                "band": "status_update",
                "statuscode": CommonService.SERVICE_STATUS_STARTING,
            },
            {
                "band": "status_update",
                "statuscode": CommonService.SERVICE_STATUS_PROCESSING,
            },
            {"band": "status_update", "statuscode": CommonService.SERVICE_STATUS_ERROR},
        ]
    )
    mock_mp.Pipe.return_value = (dummy_pipe, dummy_pipe)

    fe = workflows.frontend.Frontend(
        transport=mock.Mock(), service=service_factory, restart_service=False
    )
    with pytest.raises(workflows.Error):
        fe.run()

    service_factory.assert_called_once()
    service_process.start.assert_called_once()
    service_process.join.assert_called_once()


@mock.patch("workflows.frontend.multiprocessing")
def test_frontend_does_restart_restartable_service_on_segfault(mock_mp):
    """When the frontend is constructed with restart_service=True failing services must be restarted."""
    transport_factory = mock.Mock()
    transport1 = mock.Mock()
    transport2 = mock.Mock()
    transport3 = mock.Mock()
    transport1.return_value.connect.return_value = True
    transport2.return_value.connect.return_value = True
    transport3.return_value.connect.return_value = True
    transport_factory.side_effect = [transport1, transport2, transport3, None]
    service_factory = mock.Mock()
    service_process = mock.Mock()
    dummy_pipe = mock.Mock()
    dummy_pipe.recv.side_effect = EOFError()  # Dead on arrival
    mock_mp.Pipe.return_value = (dummy_pipe, dummy_pipe)
    mock_mp.Process.return_value = service_process

    sentinel_exception = Exception("break loop")

    service_instances = [mock.Mock(), mock.Mock()]
    service_factory.side_effect = service_instances + [sentinel_exception]

    fe = workflows.frontend.Frontend(
        transport=transport_factory, service=service_factory, restart_service=True
    )
    with pytest.raises(Exception) as e:
        fe.run()
    assert e.value == sentinel_exception

    assert service_factory.call_count == 3
    assert service_process.start.call_count == 2
    assert service_process.join.call_count == 2
    mock_mp.Process.assert_has_calls(
        [
            mock.call(args=(), kwargs=mock.ANY, target=service_instances[0].start),
            mock.call(args=(), kwargs=mock.ANY, target=service_instances[1].start),
        ],
        any_order=True,
    )
    assert service_instances[0].transport == transport2
    assert service_instances[1].transport == transport3
    assert transport1.connect.call_count == 1
    assert transport2.connect.call_count == 0
    assert transport3.connect.call_count == 0


@mock.patch("workflows.frontend.multiprocessing")
def test_frontend_does_restart_restartable_service_on_error(mock_mp):
    """When the frontend is constructed with restart_service=True failing services must be restarted."""
    sentinel_exception = Exception("break loop")
    service_instances = [mock.Mock(), mock.Mock()]
    service_factory = mock.Mock()
    service_factory.side_effect = service_instances + [sentinel_exception]
    service_process = mock.Mock()
    mock_mp.Process.return_value = service_process

    def pipe_creator(*args, **kwargs):
        """Pipe creator creates pipes."""
        dummy_pipe = MockPipe(
            [
                {
                    "band": "status_update",
                    "statuscode": CommonService.SERVICE_STATUS_NEW,
                },
                {
                    "band": "status_update",
                    "statuscode": CommonService.SERVICE_STATUS_STARTING,
                },
                {
                    "band": "status_update",
                    "statuscode": CommonService.SERVICE_STATUS_PROCESSING,
                },
                {
                    "band": "status_update",
                    "statuscode": CommonService.SERVICE_STATUS_ERROR,
                },
            ]
        )
        return (dummy_pipe, dummy_pipe)

    mock_mp.Pipe.side_effect = pipe_creator
    fe = workflows.frontend.Frontend(
        transport=mock.Mock(), service=service_factory, restart_service=True
    )
    try:
        fe.run()
        assert False, "Exception should have been raised"
    except Exception as e:
        if e != sentinel_exception:
            raise

    assert service_factory.call_count == 3
    assert service_process.start.call_count == 2
    assert service_process.join.call_count == 2
    mock_mp.Process.assert_has_calls(
        [
            mock.call(args=(), kwargs=mock.ANY, target=service_instances[0].start),
            mock.call(args=(), kwargs=mock.ANY, target=service_instances[1].start),
        ],
        any_order=True,
    )



import queue
from multiprocessing import Pipe
from unittest import mock

import pytest
from workflows.services.common_service import Commands, CommonService, Priority


def test_instantiate_basic_service():
    """Create a basic service object"""
    service = CommonService()

    assert service.get_name() is not None


def test_logging_to_frontend():
    """Log messages should be passed to frontend"""
    fe_pipe = mock.Mock()
    service = CommonService()

    # Connect service instance to pipe
    service.connect(frontend=fe_pipe)

    # Start service to initialize logging
    service.start()

    # Note that by default only warning and higher are passed to frontend
    service.log.warning(mock.sentinel.logmessage)

    fe_pipe.send.assert_called()
    assert fe_pipe.send.call_args == (({"band": "log", "payload": mock.ANY},), {})
    logrec = fe_pipe.send.call_args[0][0]["payload"]
    assert logrec.levelname == "WARNING"
    assert str(mock.sentinel.logmessage) in logrec.message


def test_adding_fieldvalue_pairs_to_log_messages():
    """Add a custom field to all outgoing log messages using the extend_log mechanism."""
    fe_pipe = mock.Mock()
    logmsg = "Test message for extension"
    service = CommonService()
    service.connect(frontend=fe_pipe)
    service.start()

    fe_pipe.send.reset_mock()
    service.log.warning(logmsg)
    fe_pipe.send.assert_called_once()
    record = fe_pipe.send.call_args[0][0]["payload"]
    assert record.message == logmsg
    assert getattr(record, "something", mock.sentinel.NA) == mock.sentinel.NA
    assert getattr(record, "furtherthing", mock.sentinel.NA) == mock.sentinel.NA

    with service.extend_log("something", "otherthing"):
        fe_pipe.send.reset_mock()
        service.log.warning(logmsg)
        fe_pipe.send.assert_called_once()
        record = fe_pipe.send.call_args[0][0]["payload"]
        assert record.message == logmsg
        assert getattr(record, "something", mock.sentinel.NA) == "otherthing"
        assert getattr(record, "furtherthing", mock.sentinel.NA) == mock.sentinel.NA

        with service.extend_log("furtherthing", "morething"):
            fe_pipe.send.reset_mock()
            service.log.warning(logmsg)
            fe_pipe.send.assert_called_once()
            record = fe_pipe.send.call_args[0][0]["payload"]
            assert record.message == logmsg
            assert getattr(record, "something", mock.sentinel.NA) == "otherthing"
            assert getattr(record, "furtherthing", mock.sentinel.NA) == "morething"

    fe_pipe.send.reset_mock()
    service.log.warning(logmsg)
    fe_pipe.send.assert_called_once()
    record = fe_pipe.send.call_args[0][0]["payload"]
    assert record.message == logmsg
    assert getattr(record, "something", mock.sentinel.NA) == mock.sentinel.NA
    assert getattr(record, "furtherthing", mock.sentinel.NA) == mock.sentinel.NA


def test_log_message_fieldvalue_pairs_are_removed_outside_their_context():
    """Custom fields added to outgoing log messages don't live on outside their extend_log context."""
    fe_pipe = mock.Mock()
    logmsg = "Test message for extension"
    service = CommonService()
    service.connect(frontend=fe_pipe)
    service.start()

    fe_pipe.send.reset_mock()
    service.log.warning(logmsg)
    fe_pipe.send.assert_called_once()
    record = fe_pipe.send.call_args[0][0]["payload"]
    assert record.message == logmsg
    assert getattr(record, "something", mock.sentinel.NA) == mock.sentinel.NA

    trace_exception = ZeroDivisionError(mock.sentinel.trace)
    try:
        with service.extend_log("something", "otherthing"):
            fe_pipe.send.reset_mock()
            service.log.warning(logmsg)
            fe_pipe.send.assert_called_once()
            record = fe_pipe.send.call_args[0][0]["payload"]
            assert record.message == logmsg
            assert getattr(record, "something", mock.sentinel.NA) == "otherthing"

            raise trace_exception
        assert False  # Getting to this point means something went horribly wrong.

    except ZeroDivisionError as e:
        assert e == trace_exception

        fe_pipe.send.reset_mock()
        service.log.warning(logmsg)
        fe_pipe.send.assert_called_once()
        record = fe_pipe.send.call_args[0][0]["payload"]
        assert record.message == logmsg
        assert getattr(record, "something", mock.sentinel.NA) == mock.sentinel.NA


def test_log_message_fieldvalue_pairs_are_attached_to_unhandled_exceptions_and_logged_properly():
    """When an exception falls through the extend_log context handler the fields are removed from future log messages,
    but they are also attached to the exception object, as they may contain valuable information for debugging."""
    fe_pipe = mock.Mock()
    service = CommonService()
    service.connect(frontend=fe_pipe)
    service.start()

    try:
        with service.extend_log("something", "otherthing"):
            nonsense = 7 / 0  # noqa
        assert False  # Getting to this point means something went horribly wrong.
    except ZeroDivisionError as e:
        assert getattr(e, "workflows_log_something", None) == "otherthing"

        # Now check that the default process_uncaught_exception handler adds that
        # field to the resulting log message
        fe_pipe.send.reset_mock()
        service.process_uncaught_exception(e)
        fe_pipe.send.assert_called_once()
        record = fe_pipe.send.call_args[0][0]["payload"]
        import pprint

        pprint.pprint(dir(record))
        assert "Unhandled service exception" in record.message
        assert "zero" in record.message.lower()
        assert getattr(record, "something", mock.sentinel.NA) == "otherthing"


@mock.patch("workflows.services.common_service.threading.Thread")
def test_receive_frontend_commands(mock_thread):
    """Check that messages via the command pipe land on the main queue."""
    cmd_pipe = mock.Mock()
    cmd_pipe.poll.side_effect = [
        True,
        True,
        True,
        AssertionError("Read past pipe lifetime"),
    ]
    cmd_pipe.recv.side_effect = [
        mock.sentinel.msg1,
        mock.sentinel.msg2,
        EOFError(),
        AssertionError("Read past pipe lifetime"),
    ]
    main_queue = mock.Mock()

    # Create service
    service = CommonService()
    service._CommonService__queue = main_queue
    service.connect(commands=cmd_pipe, frontend=mock.Mock())

    # Check that thread would be started
    service._CommonService__start_command_queue_listener()
    mock_thread.start.assert_called_once()
    assert mock_thread.daemon is True
    # Can't make assertions about the run() method, limitations of mocking

    # Now start listener in main thread
    service._CommonService__command_queue_listener()
    main_queue.put.assert_called()
    expected = [
        mock.call((Priority.COMMAND, 0, mock.sentinel.msg1), True, mock.ANY),
        mock.call((Priority.COMMAND, 1, mock.sentinel.msg2), True, mock.ANY),
    ]
    assert main_queue.put.call_args_list == expected


def test_observe_shutdown_command():
    """Receive a shutdown message via the command pipe and act on it.
    Check that status codes are updated properly."""
    start_cmd_queue_listener = mock.Mock()
    main_queue = mock.Mock()
    main_queue.get.side_effect = [
        (Priority.COMMAND, 1, {"band": "command", "payload": Commands.SHUTDOWN}),
        AssertionError("Not observing commands"),
    ]
    fe_pipe, fe_pipe_out = Pipe()

    # Create service
    service = CommonService()
    service.connect(commands=mock.Mock(), frontend=fe_pipe)
    # override class API to ensure overidden functions are called
    service.initializing = mock.Mock()
    service.in_shutdown = mock.Mock()
    # override service queue management
    service._CommonService__start_command_queue_listener = start_cmd_queue_listener
    service._CommonService__queue = main_queue

    # Check new status
    messages = []
    while fe_pipe_out.poll():
        message = fe_pipe_out.recv()
        if "statuscode" in message:
            messages.append(message["statuscode"])
    assert messages == [service.SERVICE_STATUS_NEW]

    # Start service
    service.start()

    # Check startup/shutdown sequence
    service.initializing.assert_called_once()
    service.in_shutdown.assert_called_once()
    main_queue.get.assert_called_once_with()
    messages = []
    while fe_pipe_out.poll():
        message = fe_pipe_out.recv()
        if "statuscode" in message:
            messages.append(message["statuscode"])
    assert messages == [
        service.SERVICE_STATUS_STARTING,
        service.SERVICE_STATUS_IDLE,
        service.SERVICE_STATUS_PROCESSING,
        service.SERVICE_STATUS_SHUTDOWN,
        service.SERVICE_STATUS_END,
    ]


def test_idle_timer_is_triggered():
    """Check that the idle timer callback is run if set."""
    start_cmd_queue_listener = mock.Mock()
    main_queue = mock.Mock()
    main_queue.get.side_effect = [
        queue.Empty(),
        (Priority.COMMAND, None, {"band": "command", "payload": Commands.SHUTDOWN}),
        AssertionError("Not observing commands"),
    ]
    fe_pipe, fe_pipe_out = Pipe()
    idle_trigger = mock.Mock()

    # Create service
    service = CommonService()
    service.connect(commands=mock.Mock(), frontend=fe_pipe)
    service._register_idle(mock.sentinel.idle_time, idle_trigger)

    # Override service queue management
    service._CommonService__start_command_queue_listener = start_cmd_queue_listener
    service._CommonService__queue = main_queue

    # Start service
    service.start()

    # Check trigger has been called after correct time
    idle_trigger.assert_called_once_with()
    main_queue.get.assert_called_with(True, mock.sentinel.idle_time)
    assert main_queue.get.call_count == 2

    # Check startup/shutdown sequence
    start_cmd_queue_listener.assert_called_once_with()
    messages = []
    while fe_pipe_out.poll():
        message = fe_pipe_out.recv()
        if "statuscode" in message:
            messages.append(message["statuscode"])
    assert messages == [
        service.SERVICE_STATUS_NEW,
        service.SERVICE_STATUS_STARTING,
        service.SERVICE_STATUS_IDLE,
        service.SERVICE_STATUS_TIMER,
        service.SERVICE_STATUS_IDLE,
        service.SERVICE_STATUS_PROCESSING,
        service.SERVICE_STATUS_SHUTDOWN,
        service.SERVICE_STATUS_END,
    ]


def test_callbacks_are_routed_correctly():
    """Incoming messages are routed to the correct callback functions"""
    cmd_pipe = mock.Mock()
    cmd_pipe.poll.return_value = True
    cmd_pipe.recv.side_effect = [
        {"band": mock.sentinel.band, "payload": mock.sentinel.payload},
        {"band": "command", "payload": Commands.SHUTDOWN},
        AssertionError("Stop queue reading thread"),
    ]
    fe_pipe, _ = Pipe()
    callback = mock.Mock()

    # Create service
    service = CommonService()
    service.connect(commands=cmd_pipe, frontend=fe_pipe)
    service._register(mock.sentinel.band, callback)

    # Start service
    service.start()

    # Check callback occured
    callback.assert_called_with(mock.sentinel.payload)


def test_log_unknown_band_data():
    """All unidentified messages should be logged to the frondend."""
    cmd_pipe = mock.Mock()
    cmd_pipe.poll.return_value = True
    cmd_pipe.recv.side_effect = [
        {"band": mock.sentinel.band, "payload": mock.sentinel.failure1},
        {"payload": mock.sentinel.failure2},
        {"band": "command", "payload": Commands.SHUTDOWN},
        AssertionError("Not observing commands"),
    ]
    fe_pipe, fe_pipe_out = Pipe()

    # Create service
    service = CommonService()
    service.connect(commands=cmd_pipe, frontend=fe_pipe)

    # Start service
    service.start()

    # Check startup/shutdown sequence
    messages = []
    while fe_pipe_out.poll():
        message = fe_pipe_out.recv()
        if message.get("band") == "log":
            messages.append(message.get("payload"))
    assert len(messages) == 2
    assert messages[0].name == "workflows.service"
    assert "unregistered band" in messages[0].message
    assert str(mock.sentinel.band) in messages[0].message
    assert messages[1].name == "workflows.service"
    assert "without band" in messages[1].message


def test_service_initialization_crashes_are_handled_correctly():
    """Log messages should be passed to frontend"""
    fe_pipe = mock.Mock()

    class CrashOnInitService(CommonService):
        """Helper class to test exception handling.
        This service crashes on initialization."""

        _service_name = "Crashservice 1"
        _logger_name = "workflows.service.crash_on_init"

        @staticmethod
        def initializing():
            """Crash."""
            assert False, "This crash needs to be handled"

    service = CrashOnInitService()
    service.connect(frontend=fe_pipe)
    service.start()

    fe_pipe.send.assert_called()

    # Service status should have been set to ERROR
    fe_pipe.send.assert_any_call(
        {"band": "status_update", "statuscode": service.SERVICE_STATUS_ERROR}
    )

    # Traceback should have been sent to log
    log_msgs = list(
        filter(
            lambda c: c[0][0] == {"band": "log", "payload": mock.ANY}
            and c[0][0]["payload"].levelname == "CRITICAL",
            fe_pipe.send.call_args_list,
        )
    )
    assert log_msgs, "No critical log message received"
    log = log_msgs[0][0][0]["payload"]
    assert "This crash needs to be handled" in log.exc_text
    assert "initializing" in log.exc_text
    assert "test_common_service" in log.exc_text
    assert log.workflows_exc_filename == "test_common_service.py"
    assert log.workflows_exc_funcName == "initializing"


def test_service_can_change_name_and_shut_itself_down():
    """Name changes should be passed to frontend"""
    fe_pipe = mock.Mock()

    class NameChangingService(CommonService):
        """Helper class to test name changing."""

        def initializing(self):
            """Change name."""
            self._set_name(mock.sentinel.newname)
            self._shutdown()

    service = NameChangingService()
    service.connect(frontend=fe_pipe)
    service.start()

    # Check for service name update
    fe_pipe.send.assert_any_call({"band": "set_name", "name": mock.sentinel.newname})
    # Service should have shut down cleanly
    fe_pipe.send.assert_any_call(
        {"band": "status_update", "statuscode": service.SERVICE_STATUS_END}
    )


def test_service_can_mark_itself_as_unstable():
    """A termination request should be passed to the frontend."""
    fe_pipe = mock.Mock()

    class UnstableService(CommonService):
        """Helper class to test self-termination requests."""

        def initializing(self):
            """Change name."""
            self._request_termination()

    service = UnstableService()
    service.connect(frontend=fe_pipe)
    service.start()

    # Check for service name update
    fe_pipe.send.assert_any_call({"band": "request_termination"})


def test_can_pass_environment_to_service():
    """Test that environment dictionaries can be passed to the service on construction and are available during runtime."""
    fe_pipe = mock.Mock()

    class EnvironmentPassingService(CommonService):
        """Helper class to test environment passing."""

        _service_name = "Environmentservice"
        _logger_name = "workflows.service.environment"

        def get_environment(self):
            """Make the service environment available to the caller."""
            return self._environment

    # Initialization without enviroment should still result in an available dictionary
    service = EnvironmentPassingService()
    service.connect(frontend=fe_pipe)
    assert (
        service.get_environment().get("non-existing-key-for-environment-passing-test")
        is None
    )

    # Check that passed environment is available
    sample_environment = {"environment": mock.sentinel.environment}
    service = EnvironmentPassingService(environment=sample_environment)
    service.connect(frontend=fe_pipe)
    assert service.get_environment().get("environment") == mock.sentinel.environment


def test_transport_object_can_be_injected():
    """Transport object must be stored, but not connected immediately."""
    service = CommonService()

    transport = mock.Mock()
    service.transport = transport

    assert service.transport == transport
    transport.connect.assert_not_called()


def test_transport_object_can_not_be_overwritten():
    """Transport object must be stored, but not connected immediately."""
    service = CommonService()

    transport1 = mock.Mock()
    transport2 = mock.Mock()
    service.transport = transport1
    with pytest.raises(AttributeError):
        service.transport = transport2

    assert service.transport == transport1


def test_transport_connection_is_started_on_initialization():
    """Transport object must be stored, but not connected immediately."""
    service = CommonService()
    transport = mock.Mock()
    service.transport = transport

    service.start()

    transport.connect.assert_called_once_with()


def test_main_queue_can_deal_with_being_empty():
    """Checks that the main service queue can be empty without crashing the service."""
    start_cmd_queue_listener = mock.Mock()
    main_queue = mock.Mock()
    main_queue.poll.side_effect = [False, True]
    main_queue.get.side_effect = [
        (Priority.COMMAND, None, {"band": "command", "payload": Commands.SHUTDOWN}),
        AssertionError("Not observing commands"),
    ]
    fe_pipe, fe_pipe_out = Pipe()

    # Create service
    service = CommonService()
    service.connect(commands=mock.Mock(), frontend=fe_pipe)

    # Override service queue management
    service._CommonService__start_command_queue_listener = start_cmd_queue_listener
    service._CommonService__queue = main_queue

    # Start service
    service.start()

    # Check startup/shutdown sequence
    start_cmd_queue_listener.assert_called_once_with()
    messages = []
    while fe_pipe_out.poll():
        message = fe_pipe_out.recv()
        if "statuscode" in message:
            messages.append(message["statuscode"])
    assert messages == [
        service.SERVICE_STATUS_NEW,
        service.SERVICE_STATUS_STARTING,
        service.SERVICE_STATUS_IDLE,
        service.SERVICE_STATUS_PROCESSING,
        service.SERVICE_STATUS_SHUTDOWN,
        service.SERVICE_STATUS_END,
    ]


def test_commands_are_processed_from_main_queue_before_transport():
    """Checks that the priority queue behaves as such."""
    start_cmd_queue_listener = mock.Mock()
    fe_pipe, fe_pipe_out = Pipe()

    mock_low_priority_callback = mock.Mock()

    # Create service
    service = CommonService()
    service.connect(commands=mock.Mock(), frontend=fe_pipe)
    main_queue = service._CommonService__queue
    main_queue.put((Priority.TRANSPORT, 1, (mock_low_priority_callback, None, None)))
    main_queue.put(
        (Priority.COMMAND, 1, {"band": "command", "payload": Commands.SHUTDOWN})
    )

    # Override service queue management
    service._CommonService__start_command_queue_listener = start_cmd_queue_listener

    # Start service
    service.start()

    # Service should terminate before processing the transport callback
    assert mock_low_priority_callback.called is False

    # Check startup/shutdown sequence
    start_cmd_queue_listener.assert_called_once_with()
    messages = []
    while fe_pipe_out.poll():
        message = fe_pipe_out.recv()
        if "statuscode" in message:
            messages.append(message["statuscode"])
    assert messages == [
        service.SERVICE_STATUS_NEW,
        service.SERVICE_STATUS_STARTING,
        service.SERVICE_STATUS_IDLE,
        service.SERVICE_STATUS_PROCESSING,
        service.SERVICE_STATUS_SHUTDOWN,
        service.SERVICE_STATUS_END,
    ]


def test_transport_callbacks_are_processed_from_main_queue():
    """Checks that the transport callbacks are properly processed from the main queue."""
    start_cmd_queue_listener = mock.Mock()
    mock_transport_callback = mock.Mock()
    main_queue = mock.Mock()
    main_queue.get.side_effect = [
        (
            Priority.TRANSPORT,
            1,
            (mock_transport_callback, mock.sentinel.header, mock.sentinel.message),
        ),
        (Priority.COMMAND, None, {"band": "command", "payload": Commands.SHUTDOWN}),
        AssertionError("Not observing commands"),
    ]
    fe_pipe, fe_pipe_out = Pipe()

    # Create service
    service = CommonService()
    service.connect(commands=mock.Mock(), frontend=fe_pipe)

    # Override service queue management
    service._CommonService__start_command_queue_listener = start_cmd_queue_listener
    service._CommonService__queue = main_queue

    # Start service
    service.start()

    # Service must call transport callback
    mock_transport_callback.assert_called_once_with(
        mock.sentinel.header, mock.sentinel.message
    )

    # Check startup/shutdown sequence
    start_cmd_queue_listener.assert_called_once_with()
    messages = []
    while fe_pipe_out.poll():
        message = fe_pipe_out.recv()
        if "statuscode" in message:
            messages.append(message["statuscode"])
    assert messages == [
        service.SERVICE_STATUS_NEW,
        service.SERVICE_STATUS_STARTING,
        service.SERVICE_STATUS_IDLE,
        service.SERVICE_STATUS_PROCESSING,
        service.SERVICE_STATUS_IDLE,
        service.SERVICE_STATUS_PROCESSING,
        service.SERVICE_STATUS_SHUTDOWN,
        service.SERVICE_STATUS_END,
    ]



from unittest import mock

import workflows.services
import workflows.services.sample_producer


def test_service_can_be_looked_up():
    """Attempt to look up the service by its name"""
    service_class = workflows.services.lookup("SampleProducer")
    assert service_class == workflows.services.sample_producer.SampleProducer


def test_service_registers_idle_timer():
    """Check that the service registers an idle event handler."""
    p = workflows.services.sample_producer.SampleProducer()
    mock_idlereg = mock.Mock()
    setattr(p, "_register_idle", mock_idlereg)

    p.initializing()

    mock_idlereg.assert_called_once_with(mock.ANY, p.create_message)


def test_service_produces_messages():
    """Check that the producer produces messages in the idle event handler."""
    p = workflows.services.sample_producer.SampleProducer()
    mock_transport = mock.Mock()
    setattr(p, "_transport", mock_transport)

    p.initializing()
    assert not mock_transport.send.called
    p.create_message()

    mock_transport.send.assert_called_once()

    p.create_message()

    assert mock_transport.send.call_count == 2
    calls = mock_transport.send.call_args_list
    assert calls[0][0][0] == calls[1][0][0]  # same destination
    assert calls[0][0][1] != calls[1][0][1]  # different message



from unittest import mock

import workflows
import workflows.recipe


def check_message_handling_via_unwrapper(
    callback, recipient, transport, rw_mock, allow_non_recipe
):
    """Test callback function of a recipe wrapper."""

    # This message does not contain an encoded recipe. It should be passed through directly.
    header = {"random-header": mock.sentinel.ID}
    message = mock.Mock()
    recipient.reset_mock()
    transport.reset_mock()
    callback(header, message)
    if allow_non_recipe:
        recipient.assert_called_once_with(None, header, message)
        transport.nack.assert_not_called()
    else:
        transport.nack.assert_called_once_with(header)
        recipient.assert_not_called()

    # This message does not contain an encoded recipe. It should be passed through directly.
    header = {"workflows-recipe": "False"}
    recipient.reset_mock()
    transport.reset_mock()
    callback(header, message)
    if allow_non_recipe:
        recipient.assert_called_once_with(None, header, message)
        transport.nack.assert_not_called()
    else:
        transport.nack.assert_called_once_with(header)
        recipient.assert_not_called()

    # This message contains an encoded recipe. It should be interpreted and the payload passed
    # through with a helper object for simple recipe-conformant replies.
    header = {"workflows-recipe": "True"}
    message = {
        "recipe": mock.sentinel.recipe,
        "recipe-pointer": mock.sentinel.recipe_pointer,
        "recipe-path": [],
        "environment": {
            "ID": mock.sentinel.GUID,
            "source": mock.sentinel.source,
            "timestamp": mock.sentinel.timestamp,
        },
        "payload": mock.sentinel.payload,
    }
    recipient.reset_mock()
    transport.reset_mock()
    callback(header, message)

    recipient.assert_called_once_with(rw_mock.return_value, header, message["payload"])
    rw_mock.assert_called_once_with(message=message, transport=transport)
    transport.nack.assert_not_called()


@mock.patch("workflows.recipe.RecipeWrapper", autospec=True)
def test_wrapping_a_subscription(rw_mock):
    """Test queue subscription with recipe wrapper."""
    transport, recipient = mock.Mock(), mock.Mock()

    sid = workflows.recipe.wrap_subscribe(
        transport,
        mock.sentinel.channel,
        recipient,
        mock.sentinel.irrelevant_extra_arg,
        keyword=mock.sentinel.keyword_arg,
    )

    # Channel and any extra arguments must be passed on to transport layer.
    # Callback function will obviously change.
    transport.subscribe.assert_called_once_with(
        mock.sentinel.channel,
        mock.ANY,
        mock.sentinel.irrelevant_extra_arg,
        keyword=mock.sentinel.keyword_arg,
    )
    callback = transport.subscribe.call_args[0][1]
    assert callback != recipient
    assert sid == transport.subscribe.return_value

    # Part II: Message handling via unwrapper
    check_message_handling_via_unwrapper(callback, recipient, transport, rw_mock, False)


@mock.patch("workflows.recipe.RecipeWrapper", autospec=True)
def test_wrapping_a_subscription_allowing_non_recipe_messages(rw_mock):
    """Test queue subscription with recipe wrapper allowing non-recipe messages to pass through."""
    transport, recipient = mock.Mock(), mock.Mock()

    sid = workflows.recipe.wrap_subscribe(
        transport,
        mock.sentinel.channel,
        recipient,
        mock.sentinel.irrelevant_extra_arg,
        keyword=mock.sentinel.keyword_arg,
        allow_non_recipe_messages=True,
    )

    transport.subscribe.assert_called_once()
    callback = transport.subscribe.call_args[0][1]
    assert sid == transport.subscribe.return_value

    # Part II: Message handling via unwrapper
    check_message_handling_via_unwrapper(callback, recipient, transport, rw_mock, True)


@mock.patch("workflows.recipe.RecipeWrapper", autospec=True)
def test_wrapping_a_broadcast_subscription(rw_mock):
    """Test topic subscription with recipe wrapper."""
    transport, recipient = mock.Mock(), mock.Mock()

    sid = workflows.recipe.wrap_subscribe_broadcast(
        transport,
        mock.sentinel.channel,
        recipient,
        mock.sentinel.irrelevant_extra_arg,
        keyword=mock.sentinel.keyword_arg,
    )

    # Channel and any extra arguments must be passed on to transport layer.
    # Callback function will obviously change.
    transport.subscribe_broadcast.assert_called_once_with(
        mock.sentinel.channel,
        mock.ANY,
        mock.sentinel.irrelevant_extra_arg,
        keyword=mock.sentinel.keyword_arg,
    )
    callback = transport.subscribe_broadcast.call_args[0][1]
    assert callback != recipient
    assert sid == transport.subscribe_broadcast.return_value

    # Part II: Message handling via unwrapper
    check_message_handling_via_unwrapper(callback, recipient, transport, rw_mock, False)


def test_wrapping_a_subscription_with_log_extension():
    """Test queue subscription with recipe wrapper, passing a log_extender function.
    If the recipe contains useful contextual information for log messages,
    such as a unique ID which can be used to connect all messages originating
    from the same recipe, then this information should be passed to the
    log_extender function."""
    transport, lext = mock.Mock(), mock.Mock()

    # Set up context manager mock
    lext.return_value.__enter__ = lext.enter
    lext.return_value.__exit__ = lext.exit

    def recipient(*args, **kwargs):
        """Dummy function accepting everything but must be run in log_extender context."""
        lext.enter.assert_called_once()
        lext.exit.assert_not_called()
        lext.recipient()

    sid = workflows.recipe.wrap_subscribe(
        transport, mock.sentinel.channel, recipient, log_extender=lext
    )

    # Channel and any extra arguments must be passed on to transport layer.
    # Callback function will obviously change.
    transport.subscribe.assert_called_once_with(mock.sentinel.channel, mock.ANY)
    callback = transport.subscribe.call_args[0][1]
    assert callback != recipient
    assert sid == transport.subscribe.return_value

    # Part II: Message handling

    # This message does not contain an encoded recipe. It should be passed through directly.
    header = {"random-header": mock.sentinel.ID}
    callback(header, mock.Mock())
    lext.assert_not_called()

    # This message does not contain an encoded recipe. It should be passed through directly.
    header = {"workflows-recipe": "False"}
    callback(header, mock.Mock())
    lext.assert_not_called()

    # This message contains an encoded recipe. The environment ID should be passed to the
    # log_extender context manager.
    header = {"workflows-recipe": "True"}
    message = {
        "recipe": {1: {}},
        "recipe-pointer": 1,
        "recipe-path": [],
        "environment": {
            "ID": mock.sentinel.GUID,
            "source": mock.sentinel.source,
            "timestamp": mock.sentinel.timestamp,
        },
        "payload": mock.sentinel.payload,
    }
    callback(header, message)
    lext.assert_called_once_with("recipe_ID", mock.sentinel.GUID)
    lext.enter.assert_called_once()
    lext.exit.assert_called_once()
    lext.recipient.assert_called_once()



import workflows.transport


def test_known_transports_is_a_dictionary():
    """Check transport register build in CommonTransport."""
    assert isinstance(workflows.transport.get_known_transports(), dict)


def test_load_any_transport():
    """Look up fallback transport mechanism."""
    assert workflows.transport.lookup(None)



import logging
import os
from unittest import mock

import workflows.logging


def test_obtain_exception_source():
    """Check that exception information can be obtained."""

    def crashfunction():
        """A function that crashes in a predictable manner."""
        everything = 42
        nothing = 0
        print(everything / nothing)

    try:
        crashfunction()
        assert False, "Test went past exception."
    except Exception:
        fp, fn, line_number, name, line = workflows.logging.get_exception_source()

        assert fp == __file__
        assert os.path.basename(fp) == fn
        assert name == "crashfunction"
        assert line_number == 15  # this will break if line number in this file changes
        assert line == "print(everything / nothing)"


def test_callback_handler_works_within_logging_framework():
    """Check that the callback handler can be used by Python logging
    and works as expected."""
    cbmock = mock.Mock()
    logmsg = "Test message for callback"
    log = logging.getLogger("workflows.tests.logging.callback")
    log.setLevel(logging.INFO)

    cbh = workflows.logging.CallbackHandler(cbmock)
    cbh.handleError = mock.Mock()
    log.addHandler(cbh)
    log.info(logmsg)

    cbmock.assert_called_once()
    assert cbmock.call_args == ((mock.ANY,), {})
    logrec = cbmock.call_args[0][0]
    assert isinstance(logrec, logging.LogRecord)
    assert logrec.name == "workflows.tests.logging.callback"
    assert logrec.levelname == "INFO"
    assert logrec.message == logmsg
    assert logrec.funcName.startswith("test_")
    assert not cbh.handleError.called

    # Now check that the callback handler can handle errors in the
    # callback function.
    logmsg = "Test message for error in logging"
    cbmock.side_effect = AttributeError("Some failure")

    log.info(logmsg)

    assert cbmock.call_count == 2
    assert cbmock.call_args == ((mock.ANY,), {})
    logrec = cbmock.call_args[0][0]
    assert isinstance(logrec, logging.LogRecord)
    assert logrec.message == logmsg
    cbh.handleError.assert_called_once()
    assert cbh.handleError.call_args == cbmock.call_args



import workflows.contrib.status_monitor as status_monitor
from unittest import mock


@mock.patch("workflows.contrib.status_monitor.curses")
@mock.patch("workflows.contrib.status_monitor.time")
@mock.patch("workflows.contrib.status_monitor.workflows.transport")
def test_status_monitor_connects_to_transport_layer(
    mock_transport, mock_time, mock_curses
):
    """Check that the status monitor properly connects to the transport layer and sets up relevant subscriptions."""
    mock_time.sleep.side_effect = KeyboardInterrupt()

    status_monitor.Monitor()
    mock_transport.lookup.assert_called_once_with(None)
    mock_transport.lookup.return_value.assert_called_once_with()
    mock_transport.lookup.return_value.return_value.connect.assert_called_once_with()

    status_monitor.Monitor(transport="some_transport")
    mock_transport.lookup.assert_called_with("some_transport")

    mock_trn = mock.Mock()
    mon = status_monitor.Monitor(mock_trn)
    mock_trn.assert_called_once_with()
    mock_trn.return_value.connect.assert_called_once_with()
    mock_trn.return_value.subscribe_broadcast.assert_called_once_with(
        mock.ANY, mon.update_status, retroactive=True
    )

    mon.run()

    mock_curses.wrapper.assert_called_once()
    run_method = mock_curses.wrapper.call_args[0][0]
    run_method(mock.Mock())

    mock_trn.return_value.disconnect.assert_called_once()



import os
import socket


def generate_unique_host_id():
    """Generate a unique ID, that is somewhat guaranteed to be unique among all
    instances running at the same time."""
    host = ".".join(reversed(socket.gethostname().split(".")))
    pid = os.getpid()
    return "%s.%d" % (host, pid)



from unittest import mock

import workflows.services
import workflows.services.sample_consumer


def test_service_can_be_looked_up():
    """Attempt to look up the service by its name"""
    service_class = workflows.services.lookup("SampleConsumer")
    assert service_class == workflows.services.sample_consumer.SampleConsumer


def test_service_subscribes_to_channel():
    """Check that the service subscribes to some location."""
    p = workflows.services.sample_consumer.SampleConsumer()
    mock_transport = mock.Mock()
    setattr(p, "_transport", mock_transport)

    p.initializing()

    mock_transport.subscribe.assert_called_once_with(mock.ANY, p.consume_message)


def test_service_can_consume_messages():
    """Check that the service registers an idle event handler."""
    p = workflows.services.sample_consumer.SampleConsumer()
    p.consume_message(None, mock.sentinel.message)
    p.consume_message({"some": "header"}, mock.sentinel.message)



"""
Runs the json validator on a json file and raises an exception if there is an error
Intended to be used as a hook for repositories which are developing recipes

Example of how this could be used in a .pre-commit-config.yaml file:
```
# Validate json recipes
- repo: https://github.com/DiamondLightSource/python-workflows
  hooks:
  - name: Recipe validation
    id: recipe-validation
    language: python
    files: ^recipes/
    entry: workflows.validate-recipe
```
"""

import argparse
import json
import logging
import sys

import workflows.recipe
import workflows


def validate_recipe(json_filename):
    """Reads a json file, tries to turn it into a recipe and then validates it.
    Exits on exception with non-zero error"""

    # Read in the file
    try:
        with open(json_filename) as f:
            recipe_text = f.read()
    except Exception:
        logging.exception(f"Could not read recipe from {json_filename}")
        raise

    # Turn it into a recipe and validate
    try:
        # Create a recipe object and validate it
        workflows.recipe.Recipe(recipe_text).validate()
    except json.JSONDecodeError as e:
        logging.error(f"JSON error in recipe {json_filename}:\n{e}")
        raise e
    except workflows.Error as e:
        logging.error(f"Problem in recipe {json_filename}:\n{e}")
        raise e
    except Exception as e:
        logging.error(f"Problem in recipe {json_filename}: {e}")
        raise e


def main():
    """Run the program from entry point"""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "files", nargs="+", help="file or list of files to be validated"
    )
    args = parser.parse_args()

    # Validate every file provided, keep list of fails
    failed_files = []
    for input_file in args.files:
        try:
            validate_recipe(input_file)
        except Exception:
            failed_files.append(input_file)

    # Let the user know which files had errors (summary of previous output)
    # Otherwise exit silently
    if failed_files:
        print(f"Errors found in the following recipes: {failed_files}")
        sys.exit(1)


if __name__ == "__main__":

    main()



from unittest import mock

import pytest
import workflows
from workflows.transport.common_transport import CommonTransport


def test_subscribe_unsubscribe_a_channel():
    """Public subscribe()-call should be routed to specific _subscribe().
    Unsubscribes should be routed to _unsubscribe() and handled properly."""
    ct = CommonTransport()
    ct._subscribe = mock.Mock()
    ct._unsubscribe = mock.Mock()
    mock_callback = mock.Mock()

    subid = ct.subscribe(
        mock.sentinel.channel,
        mock_callback,
        exclusive=mock.sentinel.exclusive,
        acknowledgement=mock.sentinel.ack,
        disable_mangling=True,
    )

    assert subid
    assert ct.subscription_callback(subid) == mock_callback
    ct._subscribe.assert_called_once_with(
        subid,
        mock.sentinel.channel,
        mock.ANY,
        exclusive=mock.sentinel.exclusive,
        acknowledgement=mock.sentinel.ack,
    )
    callback = ct._subscribe.call_args[0][2]
    callback(mock.sentinel.header, mock.sentinel.message)
    mock_callback.assert_called_once_with(mock.sentinel.header, mock.sentinel.message)
    # Should not be able to drop callback reference for live subscription
    with pytest.raises(workflows.Error):
        ct.drop_callback_reference(subid)

    ct.unsubscribe(subid)

    ct._unsubscribe.assert_called_once_with(subid)

    # Multiple unsubscribes should not work
    with pytest.raises(workflows.Error):
        ct.unsubscribe(subid)
    ct._unsubscribe.assert_called_once()

    # Callback should still be connected
    ct.subscription_callback(subid)

    ct.drop_callback_reference(subid)
    with pytest.raises(workflows.Error):
        ct.subscription_callback(subid)

    # Should not be able to double-drop reference
    with pytest.raises(workflows.Error):
        ct.drop_callback_reference(subid)


def test_subscription_messages_pass_mangling_function():
    """A message received via a subscription must pass through the mangle function."""
    ct = CommonTransport()
    ct._subscribe = mock.Mock()
    mock_callback = mock.Mock()

    subid = ct.subscribe(mock.sentinel.channel, mock_callback)
    assert subid
    assert ct.subscription_callback(subid) != mock_callback
    ct.subscription_callback(subid)(mock.sentinel.header, mock.sentinel.message)
    mock_callback.assert_called_once_with(mock.sentinel.header, mock.sentinel.message)


def test_simple_subscribe_unsubscribe_a_broadcast():
    """Public subscribe_bc()-call should be routed to specific _subscribe_bc().
    Unsubscribes should be routed to _unsubscribe() and handled properly."""
    ct = CommonTransport()
    ct._subscribe_broadcast = mock.Mock()
    ct._unsubscribe = mock.Mock()
    mock_callback = mock.Mock()

    subid = ct.subscribe_broadcast(
        mock.sentinel.channel,
        mock_callback,
        retroactive=mock.sentinel.retro,
        disable_mangling=True,
    )

    assert subid
    assert ct.subscription_callback(subid) == mock_callback
    ct._subscribe_broadcast.assert_called_once_with(
        subid, mock.sentinel.channel, mock.ANY, retroactive=mock.sentinel.retro
    )
    callback = ct._subscribe_broadcast.call_args[0][2]
    callback(mock.sentinel.header, mock.sentinel.message)
    mock_callback.assert_called_once_with(mock.sentinel.header, mock.sentinel.message)
    # Should not be able to drop callback reference for live subscription
    with pytest.raises(workflows.Error):
        ct.drop_callback_reference(subid)

    ct.unsubscribe(subid)

    ct._unsubscribe.assert_called_once_with(subid)
    # Multiple unsubscribes should not work
    with pytest.raises(workflows.Error):
        ct.unsubscribe(subid)
    ct._unsubscribe.assert_called_once()

    # Callback should still be connected
    ct.subscription_callback(subid)

    ct.drop_callback_reference(subid)
    with pytest.raises(workflows.Error):
        ct.subscription_callback(subid)

    # Should not be able to double-drop reference
    with pytest.raises(workflows.Error):
        ct.drop_callback_reference(subid)


def test_broadcast_subscription_messages_pass_mangling_function():
    """A message received via a broadcast subscription must pass through the mangle function."""
    ct = CommonTransport()
    ct._subscribe_broadcast = mock.Mock()
    mock_callback = mock.Mock()

    subid = ct.subscribe_broadcast(mock.sentinel.channel, mock_callback)
    assert subid
    assert ct.subscription_callback(subid) != mock_callback
    ct.subscription_callback(subid)(mock.sentinel.header, mock.sentinel.message)
    mock_callback.assert_called_once_with(mock.sentinel.header, mock.sentinel.message)


@pytest.mark.parametrize("mangling", [None, True, False])
def test_callbacks_can_be_intercepted(mangling):
    """The function called on message receipt must be interceptable."""
    ct = CommonTransport()
    ct._subscribe = mock.Mock()
    mock_true_callback = mock.Mock()
    intercept = mock.Mock()

    if mangling is None:
        subid = ct.subscribe(mock.sentinel.channel, mock_true_callback)
    else:
        subid = ct.subscribe(
            mock.sentinel.channel, mock_true_callback, disable_mangling=mangling
        )

    # Original code path
    ct.subscription_callback(subid)(mock.sentinel.header, mock.sentinel.message)
    mock_true_callback.assert_called_once_with(
        mock.sentinel.header, mock.sentinel.message
    )

    # Intercept and don't pass on
    mock_true_callback.reset_mock()
    ct.subscription_callback_set_intercept(intercept)

    ct.subscription_callback(subid)(mock.sentinel.header, mock.sentinel.message)
    assert not mock_true_callback.called
    intercept.assert_called_once_with(mock.ANY)
    intercept.return_value.assert_called_once_with(
        mock.sentinel.header, mock.sentinel.message
    )

    # Pass through (tests the value passed to the interceptor function is sensible)
    intercept = lambda x: x
    ct.subscription_callback_set_intercept(intercept)

    ct.subscription_callback(subid)(mock.sentinel.header, mock.sentinel.message)
    mock_true_callback.assert_called_once_with(
        mock.sentinel.header, mock.sentinel.message
    )

    # Disable interception
    mock_true_callback.reset_mock()
    ct.subscription_callback_set_intercept(None)

    ct.subscription_callback(subid)(mock.sentinel.header, mock.sentinel.message)
    mock_true_callback.assert_called_once_with(
        mock.sentinel.header, mock.sentinel.message
    )


def test_simple_send_message():
    """Pass messages to send(), should be routed to specific _send()"""
    ct = CommonTransport()
    ct._send = mock.Mock()

    ct.send(
        mock.sentinel.destination,
        mock.sentinel.message,
        headers=mock.sentinel.headers,
        expiration=mock.sentinel.expiration,
        transaction=mock.sentinel.transaction,
        something=mock.sentinel.something,
    )

    ct._send.assert_called_once_with(
        mock.sentinel.destination,
        mock.sentinel.message,
        headers=mock.sentinel.headers,
        expiration=mock.sentinel.expiration,
        transaction=mock.sentinel.transaction,
        something=mock.sentinel.something,
    )

    ct.send(mock.sentinel.destination, mock.sentinel.message)

    ct._send.assert_called_with(mock.sentinel.destination, mock.sentinel.message)


def test_simple_broadcast_message():
    """Pass messages to broadcast(), should be routed to specific _broadcast()"""

    ct = CommonTransport()
    ct._broadcast = mock.Mock()

    ct.broadcast(
        mock.sentinel.destination,
        mock.sentinel.message,
        headers=mock.sentinel.headers,
        expiration=mock.sentinel.expiration,
        transaction=mock.sentinel.transaction,
    )

    ct._broadcast.assert_called_once_with(
        mock.sentinel.destination,
        mock.sentinel.message,
        headers=mock.sentinel.headers,
        expiration=mock.sentinel.expiration,
        transaction=mock.sentinel.transaction,
    )

    ct.broadcast(mock.sentinel.destination, mock.sentinel.message)

    ct._broadcast.assert_called_with(mock.sentinel.destination, mock.sentinel.message)


def test_create_and_destroy_transactions():
    "Create, commit and abort transactions."
    ct = CommonTransport()
    ct._transaction_begin = mock.Mock()
    ct._transaction_commit = mock.Mock()
    ct._transaction_abort = mock.Mock()

    t = ct.transaction_begin()

    assert t
    ct._transaction_begin.assert_called_once_with(t)

    ct.transaction_abort(t)
    with pytest.raises(workflows.Error):
        ct.transaction_commit(t)
    ct._transaction_abort.assert_called_once_with(t)

    t2 = ct.transaction_begin()
    assert t2
    assert t != t2
    ct.transaction_commit(t2)
    with pytest.raises(workflows.Error):
        ct.transaction_abort(t2)
    ct._transaction_commit.assert_called_once_with(t2)


def test_messages_can_be_acknowledged_and_rejected():
    """Check that ack() and nack() calls are passed to implementations."""
    ct = CommonTransport()
    ct._ack = mock.Mock()
    ct._nack = mock.Mock()

    ct.ack(mock.sentinel.message_id1, mock.sentinel.subscription_id1)
    ct.nack(mock.sentinel.message_id2, mock.sentinel.subscription_id2)
    ct.ack({"message-id": mock.sentinel.message_id3}, mock.sentinel.subscription_id3)
    ct.nack({"message-id": mock.sentinel.message_id4}, mock.sentinel.subscription_id4)
    ct.ack(
        {
            "message-id": mock.sentinel.message_id5,
            "subscription": mock.sentinel.subscription_id5,
        }
    )
    ct.nack(
        {
            "message-id": mock.sentinel.message_id6,
            "subscription": mock.sentinel.subscription_id6,
        }
    )

    ct._ack.assert_any_call(mock.sentinel.message_id1, mock.sentinel.subscription_id1)
    ct._ack.assert_any_call(mock.sentinel.message_id3, mock.sentinel.subscription_id3)
    ct._ack.assert_any_call(mock.sentinel.message_id5, mock.sentinel.subscription_id5)
    ct._nack.assert_any_call(mock.sentinel.message_id2, mock.sentinel.subscription_id2)
    ct._nack.assert_any_call(mock.sentinel.message_id4, mock.sentinel.subscription_id4)
    ct._nack.assert_any_call(mock.sentinel.message_id6, mock.sentinel.subscription_id6)

    with pytest.raises(workflows.Error):
        ct.ack(None)
    with pytest.raises(workflows.Error):
        ct.ack(mock.sentinel.crash)
    with pytest.raises(workflows.Error):
        ct.ack({"message-id": mock.sentinel.crash})
    with pytest.raises(workflows.Error):
        ct.ack({"message-id": None, "subscription": mock.sentinel.crash})
    with pytest.raises(workflows.Error):
        ct.nack(None)
    with pytest.raises(workflows.Error):
        ct.nack(mock.sentinel.crash)
    with pytest.raises(workflows.Error):
        ct.nack({"message-id": mock.sentinel.crash})
    with pytest.raises(workflows.Error):
        ct.nack({"message-id": None, "subscription": mock.sentinel.crash})


def test_unimplemented_communication_methods_should_fail():
    """Check that low-level communication calls raise NotImplementedError when not
    overridden."""
    ct = CommonTransport()
    assert not ct.connect()
    for function, argcount in [
        ("_subscribe", 3),
        ("_subscribe_broadcast", 3),
        ("_unsubscribe", 1),
        ("_send", 2),
        ("_broadcast", 2),
        ("_ack", 2),
        ("_nack", 2),
        ("_transaction_begin", 1),
        ("_transaction_abort", 1),
        ("_transaction_commit", 1),
    ]:
        with pytest.raises(NotImplementedError):
            getattr(ct, function)(*([mock.Mock()] * argcount))



from workflows.recipe.recipe import Recipe
from workflows.recipe.wrapper import RecipeWrapper
from workflows.recipe.validate import validate_recipe

__all__ = [
    "Recipe",
    "RecipeWrapper",
    "validate_recipe",
    "wrap_subscribe",
    "wrap_subscribe_broadcast",
]


def _wrap_subscription(
    transport_layer, subscription_call, channel, callback, *args, **kwargs
):
    """Internal method to create an intercepting function for incoming messages
    to interpret recipes. This function is then used to subscribe to a channel
    on the transport layer.
      :param transport_layer: Reference to underlying transport object.
      :param subscription_call: Reference to the subscribing function of the
                                transport layer.
      :param channel:  Channel name to subscribe to.
      :param callback: Real function to be called when messages are received.
                       The callback will pass three arguments,
                       a RecipeWrapper object (details below), the header as
                       a dictionary structure, and the message.

      :param allow_non_recipe_messages: Pass on incoming messages that do not
                       include recipe information. In this case the first
                       argument to the callback function will be 'None'.
      :param log_extender: If the recipe contains useful contextual information
                       for log messages, such as a unique ID which can be used
                       to connect all messages originating from the same
                       recipe, then the information will be passed to this
                       function, which must be a context manager factory.
      :return:         Return value of call to subscription_call.
    """

    allow_non_recipe_messages = kwargs.pop("allow_non_recipe_messages", False)
    log_extender = kwargs.pop("log_extender", None)

    def unwrap_recipe(header, message):
        """This is a helper function unpacking incoming messages when they are
        in a recipe format. Other messages are passed through unmodified.
        :param header:  A dictionary of message headers. If the header contains
                        an entry 'workflows-recipe' then the message is parsed
                        and the embedded recipe information is passed on in a
                        RecipeWrapper object to the target function.
        :param message: Incoming deserialized message object.
        """
        if header.get("workflows-recipe") in (True, "True", "true", 1):
            rw = RecipeWrapper(message=message, transport=transport_layer)
            if log_extender and rw.environment and rw.environment.get("ID"):
                with log_extender("recipe_ID", rw.environment["ID"]):
                    return callback(rw, header, message.get("payload"))
            return callback(rw, header, message.get("payload"))
        if allow_non_recipe_messages:
            return callback(None, header, message)
        #   self.log.warning('Discarding non-recipe message:\n' + \
        #                    "First 1000 characters of header:\n%s\n" + \
        #                    "First 1000 characters of message:\n%s",
        #                    str(header)[:1000], str(message)[:1000])
        transport_layer.nack(header)

    return subscription_call(channel, unwrap_recipe, *args, **kwargs)


def wrap_subscribe(transport_layer, channel, callback, *args, **kwargs):
    """Listen to a queue on the transport layer, similar to the subscribe call in
    transport/common_transport.py. Intercept all incoming messages and parse
    for recipe information.
    See common_transport.subscribe for possible additional keyword arguments.
      :param transport_layer: Reference to underlying transport object.
      :param channel:  Queue name to subscribe to.
      :param callback: Function to be called when messages are received.
                       The callback will pass three arguments,
                       a RecipeWrapper object (details below), the header as
                       a dictionary structure, and the message.
      :return: A unique subscription ID
    """

    return _wrap_subscription(
        transport_layer, transport_layer.subscribe, channel, callback, *args, **kwargs
    )


def wrap_subscribe_broadcast(transport_layer, channel, callback, *args, **kwargs):
    """Listen to a topic on the transport layer, similar to the
    subscribe_broadcast call in transport/common_transport.py. Intercept all
    incoming messages and parse for recipe information.
    See common_transport.subscribe_broadcast for possible arguments.
      :param transport_layer: Reference to underlying transport object.
      :param channel:  Topic name to subscribe to.
      :param callback: Function to be called when messages are received.
                       The callback will pass three arguments,
                       a RecipeWrapper object (details below), the header as
                       a dictionary structure, and the message.
      :return: A unique subscription ID
    """

    return _wrap_subscription(
        transport_layer,
        transport_layer.subscribe_broadcast,
        channel,
        callback,
        *args,
        **kwargs
    )



import workflows.recipe


class RecipeWrapper:
    """A wrapper object which contains a recipe and a number of functions to make
    life easier for recipe users.
    """

    def __init__(
        self, message=None, transport=None, recipe=None, environment=None, **kwargs
    ):
        """Create a RecipeWrapper object from a wrapped message.
        References to the transport layer are required to send directly to
        connected downstream processes.
        """
        if message:
            self.recipe = workflows.recipe.Recipe(message["recipe"])
            self.recipe_pointer = int(message["recipe-pointer"])
            self.recipe_step = self.recipe[self.recipe_pointer]
            self.recipe_path = message.get("recipe-path", [])
            if environment is None:
                self.environment = message.get("environment", {})
            else:
                self.environment = environment
            self.payload = message.get("payload")
        elif recipe:
            if isinstance(recipe, workflows.recipe.Recipe):
                self.recipe = recipe
            else:
                self.recipe = workflows.recipe.Recipe(recipe)
            self.recipe_pointer = None
            self.recipe_step = None
            self.recipe_path = []
            self.environment = environment or {}
            self.payload = None
        else:
            raise ValueError(
                "A message or recipe is required to create a RecipeWrapper object."
            )
        self.default_channel = None
        self.transport = transport

    def send(self, *args, **kwargs):
        """Send messages to another service that is connected to the currently
        running service via the recipe. The 'send' method will either use a
        default channel name, set via the set_default_channel method, or an
        unnamed output definition.
        """
        if not self.transport:
            raise ValueError(
                "This RecipeWrapper object does not contain "
                "a reference to a transport object."
            )

        if not self.recipe_step:
            raise ValueError(
                "This RecipeWrapper object does not contain "
                "a recipe with a selected step."
            )

        if "output" not in self.recipe_step:
            # The current recipe step does not have output channels.
            return

        if isinstance(self.recipe_step["output"], dict):
            # The current recipe step does have named output channels.
            if self.default_channel:
                # Use named output channel
                self.send_to(self.default_channel, *args, **kwargs)

        else:
            # The current recipe step does have unnamed output channels.
            self._send_to_destinations(self.recipe_step["output"], *args, **kwargs)

    def send_to(self, channel, *args, **kwargs):
        """Send messages to another service that is connected to the currently
        running service via the recipe. Discard messages if the recipe does
        not have anything connected to the specified output channel.
        """
        if not self.transport:
            raise ValueError(
                "This RecipeWrapper object does not contain "
                "a reference to a transport object."
            )

        if not self.recipe_step:
            raise ValueError(
                "This RecipeWrapper object does not contain "
                "a recipe with a selected step."
            )

        if "output" not in self.recipe_step:
            # The current recipe step does not have output channels.
            return

        if not isinstance(self.recipe_step["output"], dict):
            # The current recipe step does not have named output channels.
            if self.default_channel == channel:
                # Use unnamed output channels
                self.send(*args, **kwargs)
            return

        if channel not in self.recipe_step["output"]:
            # The current recipe step does not have an output channel with this name.
            return

        self._send_to_destinations(self.recipe_step["output"][channel], *args, **kwargs)

    def set_default_channel(self, channel):
        """Define one named output channel to be equivalent to unnamed output
        channels. For this channel send() and send_to() will be identical."""
        self.default_channel = channel

    def start(self, header=None, **kwargs):
        """Trigger the start of a recipe, sending the defined payloads to the
        recipients set in the recipe. Any parameters to this function are
        passed to the transport send/broadcast methods.
        If the wrapped recipe has already been started then a ValueError will
        be raised.
        """
        if not self.transport:
            raise ValueError(
                "This RecipeWrapper object does not contain "
                "a reference to a transport object."
            )

        if self.recipe_step:
            raise ValueError("This recipe has already been started.")

        for destination, payload in self.recipe["start"]:
            self._send_to_destination(destination, header, payload, kwargs)

    def checkpoint(self, message, header=None, delay=0, **kwargs):
        """Send a message to the current recipe destination. This can be used to
        keep a state for longer processing tasks.
        :param delay: Delay transport of message by this many seconds
        """
        if not self.transport:
            raise ValueError(
                "This RecipeWrapper object does not contain "
                "a reference to a transport object."
            )

        if not self.recipe_step:
            raise ValueError(
                "This RecipeWrapper object does not contain "
                "a recipe with a selected step."
            )

        kwargs["delay"] = delay

        self._send_to_destination(
            self.recipe_pointer, header, message, kwargs, add_path_step=False
        )

    def apply_parameters(self, parameters):
        """Recursively apply parameter replacement (see recipe.py) to the wrapped
        recipe, updating internal references afterwards.
        While this operation is useful for testing it should not be used in
        production. Replacing parameters means that the recipe changes as it is
        passed down the chain of services. This makes debugging very difficult.
        """
        self.recipe.apply_parameters(parameters)
        self.recipe_step = self.recipe[self.recipe_pointer]

    def _generate_full_recipe_message(self, destination, message, add_path_step):
        """Factory function to generate independent message objects for
        downstream recipients with different destinations."""
        if add_path_step and self.recipe_pointer:
            recipe_path = self.recipe_path + [self.recipe_pointer]
        else:
            recipe_path = self.recipe_path

        return {
            "environment": self.environment,
            "payload": message,
            "recipe": self.recipe.recipe,
            "recipe-path": recipe_path,
            "recipe-pointer": destination,
        }

    def _send_to_destinations(self, destinations, message, header=None, **kwargs):
        """Send messages to a list of numbered destinations. This is an internal
        helper method used by the public 'send' methods.
        """
        if not isinstance(destinations, list):
            destinations = (destinations,)
        for destination in destinations:
            self._send_to_destination(destination, header, message, kwargs)

    def _send_to_destination(
        self, destination, header, payload, transport_kwargs, add_path_step=True
    ):
        """Helper function to send a message to a specific recipe destination."""
        if header:
            header = header.copy()
            header["workflows-recipe"] = True
        else:
            header = {"workflows-recipe": True}

        dest_kwargs = transport_kwargs.copy()
        if (
            "transport-delay" in self.recipe[destination]
            and "delay" not in transport_kwargs
        ):
            dest_kwargs["delay"] = self.recipe[destination]["transport-delay"]
        if self.recipe[destination].get("queue"):
            self.transport.send(
                self.recipe[destination]["queue"],
                self._generate_full_recipe_message(destination, payload, add_path_step),
                headers=header,
                **dest_kwargs
            )
        if self.recipe[destination].get("topic"):
            self.transport.broadcast(
                self.recipe[destination]["topic"],
                self._generate_full_recipe_message(destination, payload, add_path_step),
                headers=header,
                **dest_kwargs
            )
